"""Nested Clusters Optimization estimator."""

# Copyright (c) 2023
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# License: BSD 3 clause
# Implementation derived from:
# Riskfolio-Lib, Copyright (c) 2020-2023, Dany Cajas, Licensed under BSD 3 clause.
# scikit-learn, Copyright (c) 2007-2010 David Cournapeau, Fabian Pedregosa, Olivier
# Grisel Licensed under BSD 3 clause.

from copy import deepcopy

import numpy as np
import numpy.typing as npt
import pandas as pd
import sklearn as sk
import sklearn.base as skb
import sklearn.model_selection as skm
import sklearn.utils.parallel as skp

import skfolio.typing as skt
from skfolio.cluster import HierarchicalClustering
from skfolio.distance import BaseDistance, PearsonDistance
from skfolio.measures import RatioMeasure
from skfolio.model_selection import BaseCombinatorialCV, cross_val_predict
from skfolio.optimization._base import BaseOptimization
from skfolio.optimization.convex import MeanRisk
from skfolio.utils.tools import check_estimator, fit_single_estimator


class NestedClustersOptimization(BaseOptimization):
    """Nested Clusters Optimization estimator.

    Nested Clusters Optimization (NCO) is a portfolio optimization method developed by
    Marcos Lopez de Prado.

    It uses a distance matrix to compute clusters using a clustering algorithm (
    Hierarchical Tree Clustering, KMeans, etc..). For each cluster, the inner-cluster
    weights are computed by fitting the inner-estimator on each cluster using the whole
    training data. Then the outer-cluster weights are computed by training the
    outer-estimator using out-of-sample estimates of the inner-estimators with
    cross-validation. Finally, the final assets weights are the dot-product of the
    inner-weights and outer-weights.

    .. note ::

        The original paper uses KMeans as the clustering algorithm, minimum Variance for
        the inner-estimator and equal-weighted for the outer-estimator. Here we
        generalize it to all `sklearn` and `skfolio` clustering algorithms
        (HierarchicalClustering, KMeans, etc.), all portfolio optimizations
        (Mean-Variance, HRP, etc.) and risk measures (Variance, CVaR, etc.).
        To avoid data leakage at the outer-estimator, we use out-of-sample estimates to
        fit the outer estimator.

    Parameters
    ----------
    inner_estimator : BaseOptimization, optional
        :ref:`Optimization estimator <optimization>` used to estimate the inner-weights
        (also called intra-weights) which are the assets weights inside each cluster.
        The default `None` is to use :class:`~skfolio.optimization.MeanRisk`.

    outer_estimator : BaseOptimization, optional
        :ref:`Optimization estimator <optimization>` used to estimate the outer-weights
        (also called inter-weights) which are the weights applied to each cluster.
        The default `None` is to use :class:`~skfolio.optimization.MeanRisk`.

    distance_estimator : BaseDistance, optional
        :ref:`Distance estimator <distance>`.
        The distance estimator is used to estimate the codependence and the distance
        matrix needed for the computation of the linkage matrix.
        The default (`None`) is to use :class:`~skfolio.distance.PearsonDistance`.

    clustering_estimator : BaseEstimator, optional
        Clustering estimator. Must expose a `labels_` attribute after fitting.
        The clustering estimator is used to compute the clusters of the assets based on
        the distance matrix. The default (`None`) is to use
        :class:`~skfolio.cluster.HierarchicalClustering`.

        .. note ::

            Clustering estimators from `sklearn` are also supported. For example:
            `sklearn.cluster.KMeans`.

    cv : BaseCrossValidator | BaseCombinatorialCV | int | "ignore", optional
        Determines the cross-validation splitting strategy.
        The default (`None`) is to use the 5-fold cross validation `KFold()`.
        It is applied to the inner-estimators. Its out-of-sample outputs are used to
        train the outer-estimator.
        Possible inputs for `cv` are:

            * "ignore": no cross-validation is used (note that it will likely lead to data leakage with a high risk of overfitting)
            * Integer, to specify the number of folds in a :class:`sklearn.model_selection.KFold`
            * An object to be used as a cross-validation generator
            * An iterable yielding train, test splits
            * A :class:`~skfolio.model_selection.CombinatorialPurgedCV`

        If a `CombinatorialCV` cross-validator is used, each cluster out-of-sample
        outputs becomes a collection of multiple paths instead of one single path. The
        selected out-of-sample path among this collection of paths is chosen according
        to the `quantile` and `quantile_measure` parameters.

    n_jobs : int, optional
        The number of jobs to run in parallel for `fit` of all `estimators`.
        The value `-1` means using all processors.
        The default (`None`) means 1 unless in a `joblib.parallel_backend` context.

    quantile : float, default=0.5
        Quantile for a given measure (`quantile_measure`) of the out-of-sample
        inner-estimator paths when the `cv` parameter is a
        :class:`~skfolio.model_selection.CombinatorialPurgedCV` cross-validator.
        The default value is `0.5` corresponding to the path with the median measure.
        (see `cv`)

    quantile_measure : PerfMeasure or RatioMeasure or RiskMeasure or ExtraRiskMeasure, default=RatioMeasure.SHARPE_RATIO
        Measure used for the quantile path selection (see `quantile` and `cv`).
        The default is `RatioMeasure.SHARPE_RATIO`.

    verbose : int, default=0
        The verbosity level. The default value is `0`.

    portfolio_params :  dict, optional
        Portfolio parameters passed to the portfolio evaluated by the `predict` and
        `score` methods. If not provided, the `name` is copied from the optimization
        model and systematically passed to the portfolio.

    Attributes
    ----------
    weights_ : ndarray of shape (n_assets,)
        Weights of the assets.

    distance_estimator_ : BaseDistance
        Fitted `distance_estimator`.

    inner_estimators_ :  list[BaseOptimization]
        List of fitted `inner_estimator`. One per cluster for clusters containing more
        than one asset.

    outer_estimator_ : BaseOptimization
        Fitted `outer_estimator`.

    clustering_estimator_ : BaseEstimator
        Fitted `clustering_estimator`.

    n_features_in_ : int
        Number of assets seen during `fit`.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of assets seen during `fit`. Defined only when `X`
        has assets names that are all strings.

    References
    ----------
    .. [1]  "Building diversified portfolios that outperform out of sample",
        The Journal of Portfolio Management,
        Marcos López de Prado (2016)

    .. [2]  "A robust estimator of the efficient frontier",
        SSRN Electronic Journal,
        Marcos López de Prado (2019)

    .. [3] "Machine Learning for Asset Managers",
        Elements in Quantitative Finance. Cambridge University Press,
        Marcos López de Prado (2020)
    """

    inner_estimators_: list[BaseOptimization]
    outer_estimator_: BaseOptimization
    distance_estimator_: BaseDistance
    clustering_estimator_: skb.BaseEstimator

    def __init__(
        self,
        inner_estimator: BaseOptimization | None = None,
        outer_estimator: BaseOptimization | None = None,
        distance_estimator: BaseDistance | None = None,
        clustering_estimator: skb.BaseEstimator | None = None,
        cv: skm.BaseCrossValidator | BaseCombinatorialCV | str | int | None = None,
        quantile: float = 0.5,
        quantile_measure: skt.Measure = RatioMeasure.SHARPE_RATIO,
        n_jobs: int | None = None,
        verbose: int = 0,
        portfolio_params: dict | None = None,
    ):
        super().__init__(portfolio_params=portfolio_params)
        self.distance_estimator = distance_estimator
        self.clustering_estimator = clustering_estimator
        self.inner_estimator = inner_estimator
        self.outer_estimator = outer_estimator
        self.cv = cv
        self.quantile = quantile
        self.quantile_measure = quantile_measure
        self.n_jobs = n_jobs
        self.verbose = verbose

    def fit(
        self, X: npt.ArrayLike, y: npt.ArrayLike | None = None
    ) -> "NestedClustersOptimization":
        """Fit the Nested Clusters Optimization estimator.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
           Price returns of the assets.

        y : array-like of shape (n_observations, n_targets), optional
            Price returns of factors or a target benchmark.
            The default is `None`.

        Returns
        -------
        self : NestedClustersOptimization
            Fitted estimator.
        """
        self.distance_estimator_ = check_estimator(
            self.distance_estimator,
            default=PearsonDistance(),
            check_type=BaseDistance,
        )
        self.clustering_estimator_ = check_estimator(
            self.clustering_estimator,
            default=HierarchicalClustering(),
            check_type=skb.BaseEstimator,
        )
        self.outer_estimator_ = check_estimator(
            self.outer_estimator,
            default=MeanRisk(),
            check_type=BaseOptimization,
        )
        _inner_estimator = check_estimator(
            self.inner_estimator,
            default=MeanRisk(),
            check_type=BaseOptimization,
        )

        self.distance_estimator_.fit(X)
        distance = self.distance_estimator_.distance_
        n_assets = distance.shape[0]

        # To keep the asset_names --> used for visualisation
        if isinstance(X, pd.DataFrame):
            distance = pd.DataFrame(distance, columns=X.columns)

        # noinspection PyUnresolvedReferences
        self.clustering_estimator_.fit(distance)
        # noinspection PyUnresolvedReferences
        labels = self.clustering_estimator_.labels_
        n_clusters = max(labels) + 1
        clusters = [np.argwhere(labels == i).flatten() for i in range(n_clusters)]

        # Intra cluster weights
        # Fit the inner estimator on the whole training data. Those
        # base estimators will be used to retrieve the inner weights.
        # They are exposed publicly.
        # noinspection PyCallingNonCallable
        fitted_inner_estimators = skp.Parallel(n_jobs=self.n_jobs)(
            skp.delayed(fit_single_estimator)(
                sk.clone(_inner_estimator), X, y, indices=cluster_ids, axis=1
            )
            for cluster_ids in clusters
            if len(cluster_ids) != 1
        )
        fitted_inner_estimators = iter(fitted_inner_estimators)

        self.inner_estimators_ = []
        inner_weights = []
        for cluster_ids in clusters:
            w = np.zeros(n_assets)
            # For single assets, we don't run the inner optimization estimator.
            if len(cluster_ids) == 1:
                w[cluster_ids] = 1
            else:
                fitted_inner_estimator = next(fitted_inner_estimators)
                self.inner_estimators_.append(fitted_inner_estimator)
                w[cluster_ids] = fitted_inner_estimator.weights_
            inner_weights.append(w)
        inner_weights = np.array(inner_weights)
        assert not any(
            fitted_inner_estimators
        ), "fitted_inner_estimator iterator must be empty"

        # Outer cluster weights
        # To train the outer-estimator using the most data as possible, we use
        # a cross-validation to obtain the output of the cluster estimators.
        # To ensure that the data provided to each estimator are the same,
        # we need to set the random state of the cv if there is one and we
        # need to take a copy.
        if self.cv == "ignore":
            cv_predictions = None
            test_indices = slice(None)
        else:
            cv = skm.check_cv(self.cv)
            if hasattr(cv, "random_state") and cv.random_state is None:
                cv.random_state = np.random.RandomState()
            # noinspection PyCallingNonCallable
            cv_predictions = skp.Parallel(n_jobs=self.n_jobs)(
                skp.delayed(cross_val_predict)(
                    sk.clone(_inner_estimator),
                    X,
                    y,
                    cv=deepcopy(cv),
                    n_jobs=self.n_jobs,
                    verbose=self.verbose,
                    column_indices=cluster_ids,
                    method="predict",
                )
                for cluster_ids in clusters
                if len(cluster_ids) != 1
            )
            cv_predictions = iter(cv_predictions)
            if isinstance(self.cv, BaseCombinatorialCV):
                test_indices = slice(None)
            else:
                test_indices = np.sort(
                    np.concatenate([test for _, test in cv.split(X, y)])
                )

        # We validate and convert to numpy array only after inner-estimator fitting to
        # keep the assets names in case they are used in the estimator.
        if y is not None:
            X, y = self._validate_data(X, y)
            y_pred = y[test_indices]
        else:
            X = self._validate_data(X)
            y_pred = None

        X_pred = []
        fitted_inner_estimators = iter(self.inner_estimators_)
        for cluster_ids in clusters:
            if len(cluster_ids) == 1:
                pred = X[test_indices, cluster_ids[0]]
            else:
                if cv_predictions is None:
                    fitted_inner_estimator = next(fitted_inner_estimators)
                    pred = fitted_inner_estimator.predict(X[test_indices, cluster_ids])
                else:
                    pred = next(cv_predictions)
                    if isinstance(self.cv, BaseCombinatorialCV):
                        pred = pred.quantile(
                            measure=self.quantile_measure, q=self.quantile
                        )
            X_pred.append(np.asarray(pred))
        X_pred = np.array(X_pred).T
        if cv_predictions is None:
            assert not any(
                fitted_inner_estimators
            ), "fitted_inner_estimator iterator must be empty"
        else:
            assert not any(cv_predictions), "cv_predictions iterator must be empty"

        fit_single_estimator(self.outer_estimator_, X=X_pred, y=y_pred)
        outer_weights = self.outer_estimator_.weights_
        self.weights_ = outer_weights @ inner_weights
        return self



"""Tools module"""

# Copyright (c) 2023
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# License: BSD 3 clause
# Implementation derived from:
# scikit-learn, Copyright (c) 2007-2010 David Cournapeau, Fabian Pedregosa, Olivier
# Grisel Licensed under BSD 3 clause.

from collections.abc import Callable, Iterator
from enum import Enum
from functools import wraps
from typing import Any

import numpy as np
import numpy.typing as npt
import pandas as pd
import sklearn as sk
import sklearn.base as skb

__all__ = [
    "AutoEnum",
    "cached_property_slots",
    "cache_method",
    "input_to_array",
    "args_names",
    "format_measure",
    "bisection",
    "safe_split",
    "fit_single_estimator",
    "fit_and_predict",
    "deduplicate_names",
    "default_asset_names",
    "check_estimator",
]

GenericAlias = type(list[int])


class AutoEnum(str, Enum):
    """Base Enum class used in `skfolio`"""

    @staticmethod
    def _generate_next_value_(
        name: str, start: int, count: int, last_values: Any
    ) -> str:
        """Overriding `auto()`"""
        return name.lower()

    @classmethod
    def has(cls, value: str) -> bool:
        """Check if a value is in the Enum.

        Parameters
        ----------
        value : str
            Input value.

        Returns
        -------
        x : bool
            True if the value is in the Enum, False otherwise.
        """
        return value in cls._value2member_map_

    def __repr__(self) -> str:
        """Representation of the Enum"""
        return self.name


# noinspection PyPep8Naming
class cached_property_slots:
    """Cached property decorator for slots"""

    def __init__(self, func):
        self.func = func
        self.public_name = None
        self.private_name = None
        self.__doc__ = func.__doc__

    def __set_name__(self, owner, name):
        self.public_name = name
        self.private_name = f"_{name}"

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        if self.private_name is None:
            raise TypeError(
                "Cannot use cached_property instance without calling __set_name__"
                " on it."
            )
        try:
            value = getattr(instance, self.private_name)
        except AttributeError:
            value = self.func(instance)
            setattr(instance, self.private_name, value)
        return value

    def __set__(self, instance, owner=None):
        raise AttributeError(
            f"'{type(instance).__name__}' object attribute '{self.public_name}' is"
            " read-only"
        )

    __class_getitem__ = classmethod(GenericAlias)


def _make_key(args, kwds) -> int:
    """Make a cache key from optionally typed positional and keyword arguments"""
    key = args
    if kwds:
        for item in kwds.items():
            key += item
    return hash(key)


def cache_method(cache_name: str) -> Callable:
    """Decorator that caches class methods results into a class dictionary.

    Parameters
    ----------
    cache_name : str
        Name of the dictionary class attribute.

    Returns
    -------
    func : Callable
        Decorating function that caches class methods.
    """

    # To avoid memory leakage and proper garbage collection, self should not be part of
    # the cache key.
    # This is a known issue when we use functools.lru_cache on class methods.
    def decorating_function(method):
        @wraps(method)
        def wrapper(self, *args, **kwargs):
            func_name = method.__name__
            key = _make_key(args, kwargs)
            try:
                cache = getattr(self, cache_name)
            except AttributeError:
                raise AttributeError(
                    "You first need to create a dictionary class attribute named "
                    f"'{cache_name}'"
                ) from None
            if not isinstance(cache, dict):
                raise AttributeError(
                    f"'The cache named '{cache_name}' must be a "
                    f"dictionary, got {type(cache)}"
                )
            if func_name not in cache:
                cache[func_name] = {}
            c = cache[func_name]
            if key not in c:
                c[key] = method(self, *args, **kwargs)
            return c[key]

        return wrapper

    return decorating_function


def args_names(func: object) -> list[str]:
    """Returns the argument names of a function.

    Parameters
    ----------
    func : object
        Function.

    Returns
    -------
    args : list[str]
        The list of function arguments.
    """
    return [
        v for v in func.__code__.co_varnames[: func.__code__.co_argcount] if v != "self"
    ]


def check_estimator(
    estimator: skb.BaseEstimator | None, default: skb.BaseEstimator, check_type: Any
):
    """Check the estimator type and returns its cloned version it provided, otherwise
     return the default estimator.

    Parameters
    ----------
    estimator : BaseEstimator, optional
        Estimator.

    default : BaseEstimator
        Default estimator to return when `estimator` is `None`.

    check_type : Any
        Expected type of the estimator to check against.

    Returns
    -------
    estimator: Estimator
        The checked estimator or the default.
    """

    if estimator is None:
        return default
    if not isinstance(estimator, check_type):
        raise TypeError(f"Expected type {check_type}, got {type(estimator)}")
    return sk.clone(estimator)


def input_to_array(
    items: dict | npt.ArrayLike,
    n_assets: int,
    fill_value: Any,
    dim: int,
    assets_names: np.ndarray | None,
    name: str,
) -> np.ndarray:
    """Convert a collection of items (array-like or dictionary) into
    a numpy array and verify its shape.

    Parameters
    ----------
    items : np.ndarray | dict | list
        Items to verify and convert to array.

    n_assets : int
        Expected number of assets.
        Used to verify the shape of the converted array.

    fill_value : Any
        When `items` is a dictionary, elements that are not in `asset_names` are filled
        with `fill_value` in the converted array.

    dim : int
        Dimension of the final array.
        Possible values are `1` or `2`.

    assets_names : ndarray, optional
        Asset names used when `items` is a dictionary.

    name : str
        Name of the items used for error messages.

    Returns
    -------
    values : ndarray of shape (n_assets) for dim=1 or (n_groups, n_assets) for dim=2
        Converted array.
    """
    if dim not in [1, 2]:
        raise ValueError(f"dim must be 1 or 2, got {dim}")
    if isinstance(items, dict):
        if assets_names is None:
            raise ValueError(
                f"If `{name}` is provided as a dictionary, you must input `X` as a"
                " DataFrame with assets names in columns"
            )
        if dim == 1:
            arr = np.array([items.get(asset, fill_value) for asset in assets_names])
        else:
            # add assets and convert dict to ordered array
            arr = {}
            for asset in assets_names:
                elem = items.get(asset)
                if elem is None:
                    elem = [asset]
                elif np.isscalar(elem):
                    elem = [asset, elem]
                else:
                    elem = [asset, *elem]
                arr[asset] = elem
            arr = (
                pd.DataFrame.from_dict(arr, orient="index")
                .loc[assets_names]
                .to_numpy()
                .T
            )
    else:
        arr = np.asarray(items)

    if arr.ndim != dim:
        raise ValueError(f"`{name}` must be a {dim}D array, got a {arr.ndim}D array")

    if not isinstance(fill_value, str) and np.isnan(arr).any():
        raise ValueError(f"`{name}` contains NaN")

    if arr.shape[-1] != n_assets:
        if dim == 1:
            s = "(n_assets,)"
        else:
            s = "(n_groups, n_assets)"
        raise ValueError(
            f"`{name}` must be a of shape {s} with n_assets={n_assets}, "
            f"got {arr.shape[0]}"
        )
    return arr


def format_measure(x: float, percent: bool = False) -> str:
    """Format a measure number into a user-friendly string.

    Parameters
    ----------
    x : float
        Number to format.

    percent : bool, default=False
        If this is set to True, the number is formatted in percentage.

    Returns
    -------
    formatted : str
        Formatted string.
    """
    if np.isnan(x):
        return str(x)
    if percent:
        xn = x * 100
        f = "%"
    else:
        xn = x
        f = "f"
    if xn == 0:
        n = 0
    else:
        n = min(6, max(int(-np.log10(abs(xn))) + 2, 2))
    return "{value:{fmt}}".format(value=x, fmt=f".{n}{f}")


def bisection(x: list[np.ndarray]) -> Iterator[list[np.ndarray, np.ndarray]]:
    """Generator to bisect a list of array.

    Parameters
    ----------
    x : list[ndarray]
        A list of array.

    Yields
    ------
    arr :  Iterator[list[ndarray, ndarray]]
        Bisected array.
    """
    for e in x:
        n = len(e)
        if n > 1:
            mid = n // 2
            yield [e[0:mid], e[mid:n]]


def safe_indexing(
    X: npt.ArrayLike | pd.DataFrame, indices: npt.ArrayLike | None, axis: int = 0
):
    """
    Return rows, items or columns of X using indices.

    Parameters
    ----------
    X : array-like
        Data from which to sample rows.

    indices : array-like, optional
        Indices of rows or columns.
        The default (`None`) is to select the entire data.

    axis : int, default=0
        The axis along which `X` will be sub-sampled. `axis=0` will select
        rows while `axis=1` will select columns.

    Returns
    -------
    subset :
        Subset of X on axis 0.
    """
    if indices is None:
        return X
    if hasattr(X, "iloc"):
        return X.take(indices, axis=axis)
    if axis == 0:
        return X[indices]
    return X[:, indices]


def safe_split(
    X: npt.ArrayLike,
    y: npt.ArrayLike | None = None,
    indices: np.ndarray | None = None,
    axis: int = 0,
):
    """Create subset of dataset.

    Slice X, y according to indices for cross-validation.

    Parameters
    ----------
    X : array-like
        Data to be indexed.

    y : array-like
        Data to be indexed.

    indices : ndarray of int, optional
        Rows or columns to select from X and y.
        The default (`None`) is to select the entire data.

    axis : int, default=0
        The axis along which `X` will be sub-sampled. `axis=0` will select
        rows while `axis=1` will select columns.

    Returns
    -------
    X_subset : array-like
        Indexed data.

    y_subset : array-like
        Indexed targets.
    """

    X_subset = safe_indexing(X, indices=indices, axis=axis)
    if y is not None:
        y_subset = safe_indexing(y, indices=indices, axis=axis)
    else:
        y_subset = None
    return X_subset, y_subset


def fit_single_estimator(
    estimator: Any,
    X: npt.ArrayLike,
    y: npt.ArrayLike | None = None,
    indices: np.ndarray | None = None,
    axis: int = 0,
):
    """function used to fit an estimator within a job.

    Parameters
    ----------
    estimator : estimator object implementing 'fit' and 'predict'
        The object to use to fit the data.

    X : array-like of shape (n_observations, n_assets)
        The data to fit.

    y : array-like of shape (n_observations, n_targets), optional
        The target array if provided.

    indices : ndarray of int, optional
        Rows or columns to select from X and y.
        The default (`None`) is to select the entire data.

    axis : int, default=0
        The axis along which `X` will be sub-sampled. `axis=0` will select
        rows while `axis=1` will select columns.

    Returns
    -------
    fitted_estimator : estimator
        The fitted estimator.
    """

    X, y = safe_split(X, y, indices=indices, axis=axis)
    estimator.fit(X, y)
    return estimator


def fit_and_predict(
    estimator: Any,
    X: npt.ArrayLike,
    y: npt.ArrayLike | None,
    train: np.ndarray,
    test: np.ndarray | list[np.ndarray],
    fit_params: dict,
    method: str,
    column_indices: np.ndarray | None = None,
) -> npt.ArrayLike | list[npt.ArrayLike]:
    """Fit the estimator and predict values for a given dataset split.

    Parameters
    ----------
    estimator : estimator object implementing 'fit' and 'predict'
        The object to use to fit the data.

    X : array-like of shape (n_observations, n_assets)
        The data to fit.

    y : array-like of shape (n_observations, n_factors) or None
        The factor array if provided

    train : ndarray of int of shape (n_train_observations,)
        Indices of training samples.

    test : ndarray of int of shape (n_test_samples,) or list of ndarray
        Indices of test samples or list of indices.

    fit_params : dict
        Parameters that will be passed to ``estimator.fit``.

    method : str
        Invokes the passed method name of the passed estimator.

    column_indices : ndarray, optional
        Indices of columns to select.
        The default (`None`) is to select all columns.

    Returns
    -------
    predictions : array-like or list of array-like
        If `test` is an array, it returns the array-like result of calling
        'estimator.method' on `test`.
        Otherwise, if `test` is a list of arrays, it returns the list of array-like
        results of calling 'estimator.method' on each test set in `test`.
    """
    fit_params = fit_params if fit_params is not None else {}
    X, y = safe_split(X, y, indices=column_indices, axis=1)
    X_train, y_train = safe_split(X, y, indices=train, axis=0)
    if y_train is None:
        estimator.fit(X_train, **fit_params)
    else:
        estimator.fit(X_train, y_train, **fit_params)
    func = getattr(estimator, method)

    if isinstance(test, list):
        predictions = []
        for t in test:
            X_test, _ = safe_split(X, indices=t, axis=0)
            predictions.append(func(X_test))
    else:
        X_test, _ = safe_split(X, indices=test, axis=0)
        predictions = func(X_test)

    return predictions


def default_asset_names(n_assets: int) -> np.ndarray:
    """Default asset names are `["x0", "x1", ..., "x(n_assets - 1)"]`

    Parameters
    ----------
    n_assets : int
        Number of assets.

    Returns
    -------
    asset_names : ndarray of str
        Default assets names.
    """
    return np.asarray([f"x{i}" for i in range(n_assets)], dtype=object)


def deduplicate_names(names: npt.ArrayLike) -> list[str]:
    """Rename duplicated names by appending "_{duplicate_nb}" at the end.

    This function is inspired by the pandas function `_maybe_dedup_names`.

    Parameters
    ----------
    names : array-like of shape (n_names,)
        List of names.

    Returns
    -------
    names : list[str]
        Deduplicate names.
    """
    names = list(names)
    counts = {}
    for i, col in enumerate(names):
        cur_count = counts.get(col, 0)
        if cur_count > 0:
            names[i] = f"{col}_{cur_count}"
        counts[col] = cur_count + 1
    return names



"""Distance Estimators."""

# Author: Hugo Delatte <delatte.hugo@gmail.com>
# License: BSD 3 clause

from skfolio.distance._base import BaseDistance
from skfolio.distance._distance import (
    CovarianceDistance,
    DistanceCorrelation,
    KendallDistance,
    MutualInformation,
    NBinsMethod,
    PearsonDistance,
    SpearmanDistance,
)

__all__ = [
    "BaseDistance",
    "PearsonDistance",
    "KendallDistance",
    "SpearmanDistance",
    "CovarianceDistance",
    "DistanceCorrelation",
    "MutualInformation",
    "NBinsMethod",
]



"""Distance Estimators"""

# Copyright (c) 2023
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# License: BSD 3 clause

import numpy as np
import numpy.typing as npt
import pandas as pd
import scipy.spatial.distance as scd
import scipy.stats as sct
import sklearn.metrics as skm

from skfolio.distance._base import BaseDistance
from skfolio.moments import BaseCovariance, GerberCovariance
from skfolio.utils.stats import (
    NBinsMethod,
    cov_to_corr,
    n_bins_freedman,
    n_bins_knuth,
)
from skfolio.utils.tools import check_estimator


class PearsonDistance(BaseDistance):
    r"""Pearson Distance estimator.

    The codependence is computed from the Pearson correlation to which is applied a
    power and/or absolute transformation.
    This codependence is then used to compute the distance matrix.
    Some widely used distances are:

        * Standard angular distance = :math:`\sqrt{0.5 \times (1 - corr)}`
        * Absolute angular distance = :math:`\sqrt{1 - |corr|}`
        * Squared angular distance = :math:`\sqrt{1 - corr^2}`

    Parameters
    ----------
    absolute : bool, default=False
        If this is set to True, the absolute transformation is applied to the
        correlation matrix.

    power : float, default=1
        Exponent of the power transformation applied to the correlation matrix.

    Attributes
    ----------
    codependence_ : ndarray of shape (n_assets, n_assets)
        Codependence matrix.

    distance_ : ndarray of shape (n_assets, n_assets)
        Distance matrix.

    n_features_in_ : int
        Number of assets seen during `fit`.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of assets seen during `fit`. Defined only when `X`
        has assets names that are all strings.

    References
    ----------
    .. [1] "Building Diversified Portfolios that Outperform Out-of-Sample",
        Lòpez de Prado, Journal of Portfolio Management (2016)
    """

    def __init__(self, absolute: bool = False, power: float = 1):
        self.absolute = absolute
        self.power = power

    def fit(self, X: npt.ArrayLike, y=None) -> "PearsonDistance":
        """Fit the Pearson Distance estimator.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
            Price returns of the assets.

        y : Ignored
            Not used, present for API consistency by convention.

        Returns
        -------
        self : PearsonDistance
            Fitted estimator.
        """
        X = self._validate_data(X)
        corr = np.corrcoef(X.T)
        self.codependence_, self.distance_ = _corr_to_distance(
            corr, absolute=self.absolute, power=self.power
        )
        return self


class KendallDistance(BaseDistance):
    r"""Kendall Distance estimator.

    The codependence is computed from the Kendall correlation to which is applied a
    power and/or absolute transformation.
    This codependence is then used to compute the distance matrix.
    Some widely used distances are:

        * Standard angular distance = :math:`\sqrt{0.5 \times (1 - corr)}`
        * Absolute angular distance = :math:`\sqrt{1 - |corr|}`
        * Squared angular distance = :math:`\sqrt{1 - corr^2}`

    Parameters
    ----------
    absolute : bool, default=False
        If this is set to True, the absolute transformation is applied to the
        correlation matrix.
        The default is `False`.

    power : float, default=1
        Exponent of the power transformation applied to the correlation matrix.
        The default value is `1`.

    Attributes
    ----------
    codependence_ : ndarray of shape (n_assets, n_assets)
        Codependence matrix.

    distance_ : ndarray of shape (n_assets, n_assets)
        Distance matrix.

    n_features_in_ : int
        Number of assets seen during `fit`.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of assets seen during `fit`. Defined only when `X`
        has assets names that are all strings.

    References
    ----------
    .. [1] "Building Diversified Portfolios that Outperform Out-of-Sample",
        Lòpez de Prado, Journal of Portfolio Management (2016)
    """

    def __init__(self, absolute: bool = False, power: float = 1):
        self.absolute = absolute
        self.power = power

    def fit(self, X: npt.ArrayLike, y=None) -> "KendallDistance":
        """Fit the Kendall estimator.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
            Price returns of the assets.

        y : Ignored
            Not used, present for API consistency by convention.

        Returns
        -------
        self : KendallDistance
            Fitted estimator.
        """
        X = self._validate_data(X)
        corr = pd.DataFrame(X).corr(method="kendall").to_numpy()
        self.codependence_, self.distance_ = _corr_to_distance(
            corr, absolute=self.absolute, power=self.power
        )
        return self


class SpearmanDistance(BaseDistance):
    r"""Spearman Distance estimator.

    The codependence is computed from the Spearman correlation to which is applied a
    power and/or absolute transformation.
    This codependence is then used to compute the distance matrix.
    Some widely used distances are:

        * Standard angular distance = :math:`\sqrt{0.5 \times (1 - corr)}`
        * Absolute angular distance = :math:`\sqrt{1 - |corr|}`
        * Squared angular distance = :math:`\sqrt{1 - corr^2}`

    Parameters
    ----------
    absolute : bool, default=False
        If this is set to True, the absolute transformation is applied to the
        correlation matrix.
        The default is `False`.

    power : float, default=1
        Exponent of the power transformation applied to the correlation matrix.
        The default value is `1`.

    Attributes
    ----------
    codependence_ : ndarray of shape (n_assets, n_assets)
        Codependence matrix.

    distance_ : ndarray of shape (n_assets, n_assets)
        Distance matrix.

    n_features_in_ : int
        Number of assets seen during `fit`.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of assets seen during `fit`. Defined only when `X`
        has assets names that are all strings.

    References
    ----------
    .. [1] "Building Diversified Portfolios that Outperform Out-of-Sample",
        Lòpez de Prado, Journal of Portfolio Management (2016)
    """

    def __init__(self, absolute: bool = False, power: float = 1):
        self.absolute = absolute
        self.power = power

    def fit(self, X: npt.ArrayLike, y=None) -> "SpearmanDistance":
        """Fit the Spearman Kendall estimator.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
            Price returns of the assets.

        y : Ignored
            Not used, present for API consistency by convention.

        Returns
        -------
        self : SpearmanDistance
            Fitted estimator.
        """
        X = self._validate_data(X)
        corr = pd.DataFrame(X).corr(method="spearman").to_numpy()
        self.codependence_, self.distance_ = _corr_to_distance(
            corr, absolute=self.absolute, power=self.power
        )
        return self


class CovarianceDistance(BaseDistance):
    r"""Covariance Distance estimator.

    The codependence is computed from the correlation matrix of a chosen
    :ref:`covariance estimator <covariance_estimator>` to which is applied
    a power and/or absolute transformation.
    This codependence is then used to compute the distance matrix.
    Some widely used distances are:

        * Standard angular distance = :math:`\sqrt{0.5 \times (1 - corr)}`
        * Absolute angular distance = :math:`\sqrt{1 - |corr|}`
        * Squared angular distance = :math:`\sqrt{1 - corr^2}`

    Parameters
    ----------
    covariance_estimator : BaseCovariance, optional
       :ref:`Covariance estimator <covariance_estimator>`.
       The default (`None`) is to use :class:`~skfolio.moments.GerberCovariance`.

    absolute : bool, default=False
        If this is set to True, the absolute transformation is applied to the
        correlation matrix.
        The default is `False`.

    power : float, default=1
        Exponent of the power transformation applied to the correlation matrix.
        The default value is `1`.

    Attributes
    ----------
    codependence_ : ndarray of shape (n_assets, n_assets)
        Codependence matrix.

    distance_ : ndarray of shape (n_assets, n_assets)
        Distance matrix.

    covariance_estimator_: BaseCovariance
        Fitted `covariance_estimator`

    n_features_in_ : int
        Number of assets seen during `fit`.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of assets seen during `fit`. Defined only when `X`
        has assets names that are all strings.

    References
    ----------
    .. [1] "Building Diversified Portfolios that Outperform Out-of-Sample",
        Lòpez de Prado, Journal of Portfolio Management (2016)
    """

    covariance_estimator_: BaseCovariance

    def __init__(
        self,
        covariance_estimator: BaseCovariance | None = None,
        absolute: bool = False,
        power: float = 1,
    ):
        self.covariance_estimator = covariance_estimator
        self.absolute = absolute
        self.power = power

    def fit(self, X: npt.ArrayLike, y=None) -> "CovarianceDistance":
        """Fit the Covariance Distance estimator.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
            Price returns of the assets.

        y : Ignored
            Not used, present for API consistency by convention.

        Returns
        -------
        self : CovarianceDistance
            Fitted estimator.
        """
        # fitting estimators
        self.covariance_estimator_ = check_estimator(
            self.covariance_estimator,
            default=GerberCovariance(),
            check_type=BaseCovariance,
        )
        self.covariance_estimator_.fit(X)

        # we validate and convert to numpy after all models have been fitted to keep the
        # features names information.
        _ = self._validate_data(X)

        corr, _ = cov_to_corr(self.covariance_estimator_.covariance_)
        self.codependence_, self.distance_ = _corr_to_distance(
            corr, absolute=self.absolute, power=self.power
        )
        return self


class DistanceCorrelation(BaseDistance):
    """Distance Correlation estimator.

    Distance Correlation was introduced by Szekely [1]_ to capture non-linear
    dependencies.

    Parameters
    ----------
    threshold : float, default=0.5
        Distance correlation threshold.

    Attributes
    ----------
    codependence_ : ndarray of shape (n_assets, n_assets)
        Codependence matrix.

    distance_ : ndarray of shape (n_assets, n_assets)
        Distance matrix.

    n_features_in_ : int
        Number of assets seen during `fit`.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of assets seen during `fit`. Defined only when `X`
        has assets names that are all strings.

    References
    ----------
    .. [1] "Measuring and testing independence by correlation of distances"
        Gábor J. Szekely , 2005
    """

    def __init__(self, threshold: float = 0.5):
        self.threshold = threshold

    @staticmethod
    def _dcorr(x: np.ndarray, y: np.ndarray):
        """Calculate the distance correlation between two variables"""
        x = scd.squareform(scd.pdist(x.reshape(-1, 1)))
        y = scd.squareform(scd.pdist(y.reshape(-1, 1)))
        x = x - x.mean(axis=0)[np.newaxis, :] - x.mean(axis=1)[:, np.newaxis] + x.mean()
        y = y - y.mean(axis=0)[np.newaxis, :] - y.mean(axis=1)[:, np.newaxis] + y.mean()
        value = np.sqrt((x * y).sum()) / np.sqrt(
            np.sqrt((x**2).sum()) * np.sqrt((y**2).sum())
        )
        return value

    def fit(self, X: npt.ArrayLike, y=None) -> "DistanceCorrelation":
        """Fit the Distance Correlation estimator.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
            Price returns of the assets.

        y : Ignored
            Not used, present for API consistency by convention.

        Returns
        -------
        self : DistanceCorrelation
            Fitted estimator.
        """
        X = self._validate_data(X)
        n_assets = X.shape[1]
        corr = np.ones((n_assets, n_assets))
        # TODO: parallelize
        for i, j in zip(*np.triu_indices(n_assets, 1), strict=True):
            corr[i, j] = self._dcorr(x=X[:, i], y=X[:, j])
            corr[j, i] = corr[i, j]
        self.codependence_ = corr
        self.distance_ = np.sqrt(np.clip(1 - self.codependence_, a_min=0.0, a_max=1.0))
        return self


class MutualInformation(BaseDistance):
    r"""Mutual Information estimator.

    In information theory, the mutual information is a measure of the mutual dependence
    between variables.
    The related distance metric is called the variation of information.

    For two random variables X and Y, the mutual information I(X,Y) is defined as:

    .. math:: I(X,Y) = H(X) + H(Y) - H(X,Y)

    with H(X) and H(Y) the marginal entropies and H(X,Y) the joint entropy.

    The related distance metric known as the  variation of information is defined as:

    .. math:: d(X,Y) = H(X,Y) - I(X,Y) =  H(X) + H(Y) - 2 \times I(X,Y)

    and its normalization as:

    .. math:: D(X,Y) = \frac{d(X,Y)}{H(X,Y)} = \frac{H(X) + H(Y) - 2 \times I(X,Y)}{H(X) + H(Y) - I(X,Y)}

    Parameters
    ----------
    n_bins_method : NBinsMethod, default=NBinsMethod.FREEDMAN
        Method to compute the number of bins for the contingency matrix estimation used
        for the computation of the mutual information.
        Possible values are:

            * FREEDMAN (`default`)
            * KNUTH

    n_bins : int, optional
        Instead of using `n_bins_method`, you can directly specify the number of bins
        with `n_bins`.

    normalize : bool, default=True
        If this is set to True, the variation of information is normalized.
        The default is `True`.

    Attributes
    ----------
    codependence_ : ndarray of shape (n_assets, n_assets)
        Codependence matrix.

    distance_ : ndarray of shape (n_assets, n_assets)
        Distance matrix.

    n_features_in_ : int
        Number of assets seen during `fit`.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of features seen during `fit`. Defined only when `X` has feature
        names that are all strings.
    """

    def __init__(
        self,
        n_bins_method: NBinsMethod = NBinsMethod.FREEDMAN,
        n_bins: int | None = None,
        normalize: bool = True,
    ):
        self.n_bins_method = n_bins_method
        self.n_bins = n_bins
        self.normalize = normalize

    def fit(self, X: npt.ArrayLike, y=None) -> "MutualInformation":
        """Fit the Mutual Information estimator.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
            Price returns of the assets.

        y : Ignored
            Not used, present for API consistency by convention.

        Returns
        -------
        self : MutualInformation
            Fitted estimator.
        """
        X = self._validate_data(X)
        n_assets = X.shape[1]
        if self.n_bins is None:
            match self.n_bins_method:
                case NBinsMethod.FREEDMAN:
                    n_bins_func = n_bins_freedman
                case NBinsMethod.KNUTH:
                    n_bins_func = n_bins_knuth
                case _:
                    raise ValueError(f"n_bins_method {self.n_bins_method} is not valid")
            n_bins_list = [n_bins_func(x=X[:, i]) for i in range(n_assets)]
        else:
            n_bins_list = [self.n_bins] * n_assets

        corr = np.full((n_assets, n_assets), np.nan)
        dist = corr.copy()
        for i, j in zip(*np.triu_indices(n_assets), strict=True):
            n_bins = max(n_bins_list[i], n_bins_list[j])
            x = X[:, i]
            y = X[:, j]
            contingency = np.histogram2d(x, y, bins=n_bins)[0]
            mutual_information = skm.mutual_info_score(
                None, None, contingency=contingency
            )
            entropy_x = sct.entropy(np.histogram(x, n_bins)[0])
            entropy_y = sct.entropy(np.histogram(y, n_bins)[0])
            if self.normalize:
                corr[i, j] = mutual_information / min(entropy_x, entropy_y)
                dist[i, j] = max(
                    0.0,
                    (entropy_x + entropy_y - 2 * mutual_information)
                    / (entropy_x + entropy_y - mutual_information),
                )
            else:
                corr[i, j] = mutual_information
                dist[i, j] = max(0.0, entropy_x + entropy_y - 2 * mutual_information)
            corr[j, i] = corr[i, j]
            dist[j, i] = dist[i, j]
        self.codependence_ = corr
        self.distance_ = dist
        return self


def _corr_to_distance(
    corr: np.ndarray, absolute: bool, power: float
) -> tuple[np.ndarray, np.ndarray]:
    r"""Transform a correlation matrix to a codependence and distance matrix.

    Some widely used distances are:

        * Standard angular distance = :math:`\sqrt{0.5 \times (1 - corr)}`
        * Absolute angular distance = :math:`\sqrt{1 - |corr|}`
        * Squared angular distance = :math:`\sqrt{1 - corr^2}`


    Parameters
    ----------
    corr : ndarray of shape (n_assets, n_assets)
        Correlation matrix.

    absolute : bool
        If this is set to True, the absolute transformation is applied to the
        correlation matrix.

    power : float
        Exponent of the power transformation applied to the correlation matrix.

    Returns
    -------
    codependence, distance : tuple[np.ndarray, np.ndarray]
        Codependence and distance matrices.
    """
    bounds = np.array([-1, 0, 1])
    if absolute:
        corr = np.abs(corr)
        bounds = np.abs(bounds)
    corr = np.power(corr, power)
    bounds = np.power(bounds, power)
    scaler = 1 / (1 - min(bounds))
    distance = np.sqrt(np.clip(scaler * (1 - corr), a_min=0.0, a_max=1.0))
    return corr, distance



"""Bootstrap module."""

# Copyright (c) 2023
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# License: BSD 3 clause
# Implementation derived from:
# Riskfolio-Lib, Copyright (c) 2020-2023, Dany Cajas, Licensed under BSD 3 clause.

import numpy as np

__all__ = ["stationary_bootstrap"]


def optimal_block_size(x: np.ndarray) -> float:
    """Compute the optimal block size for a single series using Politis & White
    algorithm [1]_.

    Parameters
    ----------
    x : ndarray
        The input 1D-array.

    Returns
    -------
    value : float
        The optimal block size.

    References
    ----------
    .. [1] "Automatic Block-Length Selection for the Dependent Bootstrap".
        Politis & White (2004).

    .. [2] "Correction to Automatic Block-Length Selection for the Dependent Bootstrap".
        Patton, Politis & White (2009).
    """
    n = x.shape[0]
    eps = x - x.mean(0)
    b_max = np.ceil(min(3 * np.sqrt(n), n / 3))
    kn = max(5, int(np.log10(n)))
    m_max = int(np.ceil(np.sqrt(n))) + kn
    cv = 2 * np.sqrt(np.log10(n) / n)
    acv = np.zeros(m_max + 1)
    abs_acorr = np.zeros(m_max + 1)
    opt_m = None
    for i in range(m_max + 1):
        v1 = eps[i + 1 :] @ eps[i + 1 :]
        v2 = eps[: -(i + 1)] @ eps[: -(i + 1)]
        cross_prod = eps[i:] @ eps[: n - i]
        acv[i] = cross_prod / n
        abs_acorr[i] = np.abs(cross_prod) / np.sqrt(v1 * v2)
        if i >= kn:
            if np.all(abs_acorr[i - kn : i] < cv) and opt_m is None:
                opt_m = i - kn
    m = 2 * max(opt_m, 1) if opt_m is not None else m_max
    m = min(m, m_max)
    g = 0.0
    lr_acv = acv[0]
    for k in range(1, m + 1):
        lam = 1 if k / m <= 1 / 2 else 2 * (1 - k / m)
        g += 2 * lam * k * acv[k]
        lr_acv += 2 * lam * acv[k]
    d = 2 * lr_acv**2
    b = ((2 * g**2) / d) ** (1 / 3) * n ** (1 / 3)
    b = min(b, b_max)
    return b


def stationary_bootstrap(
    returns: np.ndarray,
    n_bootstrap_samples: int,
    block_size: float | None = None,
    seed: int | None = None,
) -> np.ndarray:
    """Creates `n_bootstrap_samples` samples from a multivariate return series via
    stationary bootstrapping.

    Parameters
    ----------
    returns: ndarray of shape (n_observations, n_assets)
        The returns array.

    n_bootstrap_samples: int
        The number of bootstrap samples to generate.

    block_size: float, optional
        The block size.
        If this is set to None, we estimate the optimal block size using Politis &
        White algorithm for all individual asset and the median.

    seed: int, optional
        Random seed used to initialize the pseudo-random number generator

    Returns
    -------
    value: ndarray
           The sample returns of shape (reps, nb observations, nb assets)

    """
    np.random.seed(seed=seed)
    n_observations, n_assets = returns.shape
    x = np.vstack((returns, returns))
    # Loop over reps bootstraps
    if block_size is None:
        block_size = np.median(
            [optimal_block_size(returns[:, i]) for i in range(n_assets)]
        )

    indices = np.random.randint(
        n_observations, size=(n_bootstrap_samples, n_observations)
    )
    cond = np.random.rand(n_bootstrap_samples, n_observations) >= 1.0 / block_size
    # TODO: don't use loop
    for i in range(n_bootstrap_samples):
        for j in range(1, n_observations):
            if cond[i, j]:
                indices[i, j] = indices[i, j - 1] + 1
    indices[indices > 2 * n_observations] = 0
    return x[indices, :]



"""Expected returns module."""

from skfolio.moments.expected_returns._base import (
    BaseMu,
)
from skfolio.moments.expected_returns._expected_returns import (
    EWMu,
    EmpiricalMu,
    EquilibriumMu,
    ShrunkMu,
    ShrunkMuMethods,
)

__all__ = [
    "BaseMu",
    "EmpiricalMu",
    "EWMu",
    "ShrunkMu",
    "EquilibriumMu",
    "ShrunkMuMethods",
]



from skfolio.optimization._base import BaseOptimization
from skfolio.optimization.cluster import (
    BaseHierarchicalOptimization,
    HierarchicalEqualRiskContribution,
    HierarchicalRiskParity,
    NestedClustersOptimization,
)
from skfolio.optimization.convex import (
    ConvexOptimization,
    DistributionallyRobustCVaR,
    MaximumDiversification,
    MeanRisk,
    ObjectiveFunction,
    RiskBudgeting,
)
from skfolio.optimization.ensemble import BaseComposition, StackingOptimization
from skfolio.optimization.naive import EqualWeighted, InverseVolatility, Random

__all__ = [
    "BaseOptimization",
    "InverseVolatility",
    "EqualWeighted",
    "Random",
    "ObjectiveFunction",
    "ConvexOptimization",
    "MeanRisk",
    "RiskBudgeting",
    "DistributionallyRobustCVaR",
    "MaximumDiversification",
    "BaseHierarchicalOptimization",
    "HierarchicalRiskParity",
    "HierarchicalEqualRiskContribution",
    "NestedClustersOptimization",
    "BaseComposition",
    "StackingOptimization",
]



"""Hierarchical Risk Parity Optimization estimator."""

# Copyright (c) 2023
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# License: BSD 3 clause
# The risk measure generalization and constraint features are derived
# from Riskfolio-Lib, Copyright (c) 2020-2023, Dany Cajas, Licensed under BSD 3 clause.

import numpy as np
import numpy.typing as npt
import pandas as pd
import scipy.cluster.hierarchy as sch

import skfolio.typing as skt
from skfolio.cluster import HierarchicalClustering
from skfolio.distance import BaseDistance, PearsonDistance
from skfolio.measures import ExtraRiskMeasure, RiskMeasure
from skfolio.optimization.cluster.hierarchical._base import (
    BaseHierarchicalOptimization,
)
from skfolio.prior import BasePrior, EmpiricalPrior
from skfolio.utils.tools import bisection, check_estimator


class HierarchicalRiskParity(BaseHierarchicalOptimization):
    r"""Hierarchical Risk Parity estimator.

    Hierarchical Risk Parity is a portfolio optimization method developed by Marcos
    Lopez de Prado [2]_.

    This algorithm uses a distance matrix to compute hierarchical clusters using the
    Hierarchical Tree Clustering algorithm. It then employs seriation to rearrange the
    assets in the dendrogram, minimizing the distance between leafs.

    The final step is the recursive bisection where each cluster is split between two
    sub-clusters by starting with the topmost cluster and traversing in a top-down
    manner. For each sub-cluster, we compute the total cluster risk of an inverse-risk
    allocation. A weighting factor is then computed from these two sub-cluster risks,
    which is used to update the cluster weight.

    .. note ::
        The original paper uses the variance as the risk measure and the single-linkage
        method for the Hierarchical Tree Clustering algorithm. Here we generalize it to
        multiple risk measures and linkage methods.
        The default linkage method is set to the Ward
        variance minimization algorithm, which is more stable and has better properties
        than the single-linkage method [4]_.

    Parameters
    ----------
    risk_measure : RiskMeasure or ExtraRiskMeasure, default=RiskMeasure.VARIANCE
        :class:`~skfolio.meta.RiskMeasure` or :class:`~skfolio.meta.ExtraRiskMeasure`
        of the optimization.
        Can be any of:

            * MEAN_ABSOLUTE_DEVIATION
            * FIRST_LOWER_PARTIAL_MOMENT
            * VARIANCE
            * SEMI_VARIANCE
            * CVAR
            * EVAR
            * WORST_REALIZATION
            * CDAR
            * MAX_DRAWDOWN
            * AVERAGE_DRAWDOWN
            * EDAR
            * ULCER_INDEX
            * GINI_MEAN_DIFFERENCE_RATIO
            * VALUE_AT_RISK
            * DRAWDOWN_AT_RISK
            * ENTROPIC_RISK_MEASURE
            * FOURTH_CENTRAL_MOMENT
            * FOURTH_LOWER_PARTIAL_MOMENT
            * SKEW
            * KURTOSIS

        The default is `RiskMeasure.VARIANCE`.

    prior_estimator : BasePrior, optional
        :ref:`Prior estimator <prior>`.
        The prior estimator is used to estimate the :class:`~skfolio.prior.PriorModel`
        containing the estimation of assets expected returns, covariance matrix and
        returns. The moments and returns estimations are used for the risk computation
        and the returns estimation are used by the distance matrix estimator.
        The default (`None`) is to use :class:`~skfolio.prior.EmpiricalPrior`.

    distance_estimator : BaseDistance, optional
        :ref:`Distance estimator <distance>`.
        The distance estimator is used to estimate the codependence and the distance
        matrix needed for the computation of the linkage matrix.
        The default (`None`) is to use :class:`~skfolio.distance.PearsonDistance`.

    hierarchical_clustering_estimator : HierarchicalClustering, optional
        :ref:`Hierarchical Clustering estimator <hierarchical_clustering>`.
        The hierarchical clustering estimator is used to compute the linkage matrix
        and the hierarchical clustering of the assets based on the distance matrix.
        The default (`None`) is to use
        :class:`~skfolio.cluster.HierarchicalClustering`.

    min_weights : float | dict[str, float] | array-like of shape (n_assets, ), default=0.0
        Minimum assets weights (weights lower bounds). Negative weights are not allowed.
        If a float is provided, it is applied to each asset. `None` is equivalent to
        `-np.Inf` (no lower bound). If a dictionary is provided, its (key/value) pair
        must be the (asset name/asset minium weight) and the input `X` of the `fit`
        methods must be a DataFrame with the assets names in columns. When using a
        dictionary, assets values that are not provided are assigned a minimum weight
        of `0.0`. The default is 0.0 (no short selling).

        Example:

           * min_weights = 0 --> long only portfolio (no short selling).
           * min_weights = None --> no lower bound (same as `-np.Inf`).
           * min_weights = {"SX5E": 0, "SPX": 0.1}
           * min_weights = [0, 0.1]

    max_weights : float | dict[str, float] | array-like of shape (n_assets, ), default=1.0
        Maximum assets weights (weights upper bounds). Weights above 1.0 are not
        allowed. If a float is provided, it is applied to each asset. `None` is
        equivalent to `+np.Inf` (no upper bound). If a dictionary is provided, its
        (key/value) pair must be the (asset name/asset maximum weight) and the input `X`
        of the `fit` method must be a DataFrame with the assets names in columns. When
        using a dictionary, assets values that are not provided are assigned a minimum
        weight of `1.0`. The default is 1.0 (each asset is below 100%).

        Example:

           * max_weights = 0 --> no long position (short only portfolio).
           * max_weights = 0.5 --> each weight must be below 50%.
           * max_weights = {"SX5E": 1, "SPX": 0.25}
           * max_weights = [1, 0.25]

    transaction_costs : float | dict[str, float] | array-like of shape (n_assets, ), default=0.0
        Transaction costs of the assets. It is used to add linear transaction costs to
        the optimization problem:

        .. math:: total\_cost = \sum_{i=1}^{N} c_{i} \times |w_{i} - w\_prev_{i}|

        with :math:`c_{i}` the transaction cost of asset i, :math:`w_{i}` its weight
        and :math:`w\_prev_{i}` its previous weight (defined in `previous_weights`).
        The float :math:`total\_cost` is impacting the portfolio expected return in the optimization:

        .. math:: expected\_return = \mu^{T} \cdot w - total\_cost

        with :math:`\mu` the vector af assets' expected returns and :math:`w` the
        vector of assets weights.

        If a float is provided, it is applied to each asset.
        If a dictionary is provided, its (key/value) pair must be the
        (asset name/asset cost) and the input `X` of the `fit` method must be a
        DataFrame with the assets names in columns.
        The default value is `0.0`.

        .. warning::

            Based on the above formula, the periodicity of the transaction costs
            needs to be homogenous to the periodicity of :math:`\mu`. For example, if
            the input `X` is composed of **daily** returns, the `transaction_costs` need
            to be expressed as **daily** costs.
            (See :ref:`sphx_glr_auto_examples_1_mean_risk_plot_6_transaction_costs.py`)

    management_fees : float | dict[str, float] | array-like of shape (n_assets, ), default=0.0
        Management fees of the assets. It is used to add linear management fees to the
        optimization problem:

        .. math:: total\_fee = \sum_{i=1}^{N} f_{i} \times w_{i}

        with :math:`f_{i}` the management fee of asset i and :math:`w_{i}` its weight.
        The float :math:`total\_fee` is impacting the portfolio expected return in the optimization:

        .. math:: expected\_return = \mu^{T} \cdot w - total\_fee

        with :math:`\mu` the vector af assets expected returns and :math:`w` the vector
        of assets weights.

        If a float is provided, it is applied to each asset.
        If a dictionary is provided, its (key/value) pair must be the
        (asset name/asset fee) and the input `X` of the `fit` method must be a
        DataFrame with the assets names in columns.
        The default value is `0.0`.

        .. warning::

            Based on the above formula, the periodicity of the management fees needs to
            be homogenous to the periodicity of :math:`\mu`. For example, if the input
            `X` is composed of **daily** returns, the `management_fees` need to be
            expressed in **daily** fees.

        .. note::

            Another approach is to directly impact the management fees to the input `X`
            in order to express the returns net of fees. However, when estimating the
            :math:`\mu` parameter using for example Shrinkage estimators, this approach
            would mix a deterministic value with an uncertain one leading to unwanted
            bias in the management fees.

    previous_weights : float | dict[str, float] | array-like of shape (n_assets, ), optional
        Previous weights of the assets. Previous weights are used to compute the
        portfolio total cost. If a float is provided, it is applied to each asset.
        If a dictionary is provided, its (key/value) pair must be the
        (asset name/asset previous weight) and the input `X` of the `fit` method must
        be a DataFrame with the assets names in columns.
        The default (`None`) means no previous weights.

    portfolio_params :  dict, optional
        Portfolio parameters passed to the portfolio evaluated by the `predict` and
        `score` methods. If not provided, the `name`, `transaction_costs`,
        `management_fees`, `previous_weights` and `risk_free_rate` are copied from the
        optimization model and passed to the portfolio.

    Attributes
    ----------
    weights_ : ndarray of shape (n_assets,)
        Weights of the assets.

    distance_estimator_ : BaseDistance
        Fitted `distance_estimator`.

    hierarchical_clustering_estimator_ : HierarchicalClustering
        Fitted `hierarchical_clustering_estimator`.

    n_features_in_ : int
        Number of assets seen during `fit`.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of assets seen during `fit`. Defined only when `X`
        has assets names that are all strings.

    References
    ----------
    .. [1] "Building diversified portfolios that outperform out of sample",
        The Journal of Portfolio Management,
        Marcos López de Prado (2016).

    .. [2] "A robust estimator of the efficient frontier",
        SSRN Electronic Journal,
        Marcos López de Prado (2019).

    .. [3] "Machine Learning for Asset Managers",
        Elements in Quantitative Finance. Cambridge University Press,
        Marcos López de Prado (2020).

    .. [4] "A review of two decades of correlations, hierarchies, networks and
        clustering in financial markets",
        Gautier Marti, Frank Nielsen, Mikołaj Bińkowski, Philippe Donnat (2020).
    """

    def __init__(
        self,
        risk_measure: RiskMeasure | ExtraRiskMeasure = RiskMeasure.VARIANCE,
        prior_estimator: BasePrior | None = None,
        distance_estimator: BaseDistance | None = None,
        hierarchical_clustering_estimator: HierarchicalClustering | None = None,
        min_weights: skt.MultiInput | None = 0.0,
        max_weights: skt.MultiInput | None = 1.0,
        transaction_costs: skt.MultiInput = 0.0,
        management_fees: skt.MultiInput = 0.0,
        previous_weights: skt.MultiInput | None = None,
        portfolio_params: dict | None = None,
    ):
        super().__init__(
            risk_measure=risk_measure,
            prior_estimator=prior_estimator,
            distance_estimator=distance_estimator,
            hierarchical_clustering_estimator=hierarchical_clustering_estimator,
            min_weights=min_weights,
            max_weights=max_weights,
            transaction_costs=transaction_costs,
            management_fees=management_fees,
            previous_weights=previous_weights,
            portfolio_params=portfolio_params,
        )

    def fit(self, X: npt.ArrayLike, y: None = None) -> "HierarchicalRiskParity":
        """Fit the Hierarchical Risk Parity Optimization estimator.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
            Price returns of the assets.

        y : Ignored
            Not used, present for API consistency by convention.

        Returns
        -------
        self : HierarchicalRiskParity
            Fitted estimator.
        """
        # Validate
        if not isinstance(self.risk_measure, RiskMeasure | ExtraRiskMeasure):
            raise TypeError(
                "`risk_measure` must be of type `RiskMeasure` or `ExtraRiskMeasure`"
            )
        self.prior_estimator_ = check_estimator(
            self.prior_estimator,
            default=EmpiricalPrior(),
            check_type=BasePrior,
        )
        self.distance_estimator_ = check_estimator(
            self.distance_estimator,
            default=PearsonDistance(),
            check_type=BaseDistance,
        )
        self.hierarchical_clustering_estimator_ = check_estimator(
            self.hierarchical_clustering_estimator,
            default=HierarchicalClustering(),
            check_type=HierarchicalClustering,
        )

        # Fit the estimators
        self.prior_estimator_.fit(X, y)
        prior_model = self.prior_estimator_.prior_model_
        returns = prior_model.returns

        # To keep the asset_names
        if isinstance(X, pd.DataFrame):
            returns = pd.DataFrame(returns, columns=X.columns)

        self.distance_estimator_.fit(returns)
        distance = self.distance_estimator_.distance_

        # To keep the asset_names
        if isinstance(X, pd.DataFrame):
            distance = pd.DataFrame(distance, columns=X.columns)

        self.hierarchical_clustering_estimator_.fit(distance)

        X = self._validate_data(X)
        n_assets = X.shape[1]

        min_weights, max_weights = self._convert_weights_bounds(n_assets=n_assets)
        assets_risks = self._unitary_risks(prior_model=prior_model)

        ordered_linkage_matrix = sch.optimal_leaf_ordering(
            self.hierarchical_clustering_estimator_.linkage_matrix_,
            self.hierarchical_clustering_estimator_.condensed_distance_,
        )
        sorted_assets = sch.leaves_list(ordered_linkage_matrix)

        weights = np.ones(n_assets)
        items = [sorted_assets]

        while len(items) > 0:
            new_items = []
            for clusters_ids in bisection(items):
                new_items += clusters_ids
                risks = []
                for ids in clusters_ids:
                    inv_risk_w = np.zeros(n_assets)
                    inv_risk_w[ids] = 1 / assets_risks[ids]
                    inv_risk_w /= inv_risk_w.sum()
                    risks.append(
                        self._risk(weights=inv_risk_w, prior_model=prior_model)
                    )
                left_risk, right_risk = risks
                left_cluster, right_cluster = clusters_ids
                alpha = 1 - left_risk / (left_risk + right_risk)
                # Weights constraints
                alpha = self._apply_weight_constraints_to_alpha(
                    alpha=alpha,
                    weights=weights,
                    max_weights=max_weights,
                    min_weights=min_weights,
                    left_cluster=left_cluster,
                    right_cluster=right_cluster,
                )
                weights[left_cluster] *= alpha
                weights[right_cluster] *= 1 - alpha
            items = new_items

        self.weights_ = weights
        return self



"""Base Uncertainty estimator"""

# Copyright (c) 2023
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# License: BSD 3 clause

from abc import ABC, abstractmethod
from dataclasses import dataclass

import numpy as np
import numpy.typing as npt
import sklearn.base as skb


# frozen=True with eq=False will lead to an id-based hashing which is needed for
# caching CVX models in Optimization without impacting performance
@dataclass(frozen=True, eq=False)
class UncertaintySet:
    r"""Ellipsoidal uncertainty set dataclass.

    An ellipsoidal uncertainty set is defined by its size :math:`\kappa` and
    shape :math:`S`. Ellipsoidal uncertainty set can be used with both expected returns
    and covariance:

    Expected returns ellipsoidal uncertainty set:

    .. math:: U_{\mu}=\left\{\mu\,|\left(\mu-\hat{\mu}\right)S^{-1}\left(\mu-\hat{\mu}\right)^{T}\leq\kappa^{2}\right\}

    Covariance ellipsoidal uncertainty set:

    .. math:: U_{\Sigma}=\left\{\Sigma\,|\left(\text{vec}(\Sigma)-\text{vec}(\hat{\Sigma})\right)S^{-1}\left(\text{vec}(\Sigma)-\text{vec}(\hat{\Sigma})\right)^{T}\leq k^{2}\,,\,\Sigma\succeq 0\right\}

    Attributes
    ----------
    k : float
        Size of the ellipsoid  :math:`\kappa` that defines the confidence region

    sigma : ndarray of shape (n_assets)
        Shape of the ellipsoid :math:`S`
    """

    k: float
    sigma: np.ndarray


class BaseMuUncertaintySet(skb.BaseEstimator, ABC):
    """Base class for all Mu Uncertainty Set estimators in `skfolio`.

    Notes
    -----
    All estimators should specify all the parameters that can be set
    at the class level in their ``__init__`` as explicit keyword
    arguments (no ``*args`` or ``**kwargs``).
    """

    uncertainty_set_: UncertaintySet

    @abstractmethod
    def __init__(self):
        pass

    @abstractmethod
    def fit(self, X: npt.ArrayLike, y=None):
        pass


class BaseCovarianceUncertaintySet(skb.BaseEstimator, ABC):
    """Base class for all Covariance Uncertainty Set estimators in `skfolio`.

    Notes
    -----
    All estimators should specify all the parameters that can be set
    at the class level in their ``__init__`` as explicit keyword
    arguments (no ``*args`` or ``**kwargs``).
    """

    uncertainty_set_: UncertaintySet

    @abstractmethod
    def __init__(self):
        pass

    @abstractmethod
    def fit(self, X: npt.ArrayLike, y=None):
        pass

    def _validate_X_y(self, X: npt.ArrayLike, y: npt.ArrayLike | None = None):
        """Validate X and y if provided.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
            Price returns of the assets.

        y : array-like of shape (n_observations, n_targets), optional
            Price returns of factors or a target benchmark.
            The default is `None`.

        Returns
        -------
        X : ndarray of shape (n_observations, n_assets)
            Validated price returns of the assets.
        y : ndarray of shape (n_observations, n_targets), optional
            Validated price returns of factors or a target benchmark if provided.
        """
        if y is None:
            X = self._validate_data(X)
        else:
            X, y = self._validate_data(X, y, multi_output=True)
        return X, y



"""pre-selection estimators module"""

# Copyright (c) 2023
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# License: BSD 3 clause

import numpy as np
import numpy.typing as npt
import sklearn.base as skb
import sklearn.feature_selection as skf
import sklearn.utils.validation as skv

import skfolio.typing as skt
from skfolio.measures import RatioMeasure
from skfolio.population import Population
from skfolio.portfolio import Portfolio


class DropCorrelated(skf.SelectorMixin, skb.BaseEstimator):
    """Transformer for dropping highly correlated assets.

    Simply removing all correlation pairs above the threshold will remove more assets
    than necessary and a naive sequential removal is suboptimal and depends on the
    initial assets ordering.

    Let's suppose X,Y,Z are three random variables with corr(X,Y) and corr(X,Z) above
    the threshold and corr(Y,Z) below.
    The first approach would remove X,Y,Z and the second approach would remove either
    Y and Z or X depending on the initial ordering.

    To avoid these shortcomings, we implement the below algorithm:

        * Step 1: select all correlation pairs above the threshold.
        * Step 2: sort all the selected correlation pairs from highest to lowest.
        * Step 3: for each pair, if none of the two assets has been removed, keep the
          asset with the lowest average correlation against the other assets.

    Parameters
    ----------
    threshold : float, default=0.95
        Correlation threshold. The default value is `0.95`.

    absolute : bool, default=False
        If this is set to True, we take the absolute value of the correlation. This has
        for effect to also include negatively correlated assets.

    Attributes
    ----------
    to_keep_ : ndarray of shape (n_assets, )
        Boolean array indicating which assets are remaining.

    n_features_in_ : int
        Number of assets seen during `fit`.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of assets seen during `fit`. Defined only when `X`
        has assets names that are all strings.
    """

    to_keep_: np.ndarray

    def __init__(self, threshold: float = 0.95, absolute: bool = False):
        self.threshold = threshold
        self.absolute = absolute

    def fit(self, X: npt.ArrayLike, y=None):
        """Run the correlation transformer and get the appropriate assets.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
            Price returns of the assets.

        y : Ignored
            Not used, present for API consistency by convention.

        Returns
        -------
        self : DropCorrelated
            Fitted estimator.
        """
        X = self._validate_data(X)
        if not -1 <= self.threshold <= 1:
            raise ValueError("`threshold` must be between -1 and 1")

        n_assets = X.shape[1]
        corr = np.corrcoef(X.T)
        mean_corr = corr.mean(axis=0)

        triu_idx = np.triu_indices(n_assets, 1)

        # select all correlation pairs above the threshold
        selected_idx = np.argwhere(corr[triu_idx] > self.threshold).flatten()

        # sort all the selected correlation pairs from highest to lowest
        selected_idx = selected_idx[np.argsort(-corr[triu_idx][selected_idx])]

        # for each pair, if none of the two assets has been removed, keep the asset with
        # the lowest average correlation with other assets
        to_remove = set()
        for idx in selected_idx:
            i, j = triu_idx[0][idx], triu_idx[1][idx]
            if i not in to_remove and j not in to_remove:
                if mean_corr[i] > mean_corr[j]:
                    to_remove.add(i)
                else:
                    to_remove.add(j)
        self.to_keep_ = ~np.isin(np.arange(n_assets), list(to_remove))
        return self

    def _get_support_mask(self):
        skv.check_is_fitted(self)
        return self.to_keep_


class SelectKExtremes(skf.SelectorMixin, skb.BaseEstimator):
    """Transformer for selecting the `k` best or worst assets.

    Keep the `k` best or worst assets according to a given measure.

    Parameters
    ----------
    k : int, default=10
        Number of assets to select. If `k` is higher than the number of assets, all
        assets are selected.

    measure : Measure, default=RatioMeasure.SHARPE_RATIO
        The :ref:`measure <measures_ref>` used to sort the assets.
        The default is `RatioMeasure.SHARPE_RATIO`.

    highest : bool, default=True
        If this is set to True, the `k` assets with the highest `measure` are selected,
        otherwise it is the `k` lowest.

    Attributes
    ----------
    to_keep_ : ndarray of shape (n_assets, )
       Boolean array indicating which assets are remaining.

    n_features_in_ : int
       Number of assets seen during `fit`.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
       Names of features seen during `fit`. Defined only when `X`
       has feature names that are all strings.
    """

    to_keep_: np.ndarray

    def __init__(
        self,
        k: int = 10,
        measure: skt.Measure = RatioMeasure.SHARPE_RATIO,
        highest: bool = True,
    ):
        self.k = k
        self.measure = measure
        self.highest = highest

    def fit(self, X: npt.ArrayLike, y=None) -> "SelectKExtremes":
        """Run the SelectKExtremes transformer and get the appropriate assets.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
            Price returns of the assets.

        y : Ignored
            Not used, present for API consistency by convention.

        Returns
        -------
        self : SelectKExtremes
            Fitted estimator.
        """
        X = self._validate_data(X)
        k = int(self.k)
        if k <= 0:
            raise ValueError("`k` must be strictly positive")
        n_assets = X.shape[1]
        # Build a population of single assets portfolio
        population = Population([])
        for i in range(n_assets):
            weights = np.zeros(n_assets)
            weights[i] = 1
            population.append(Portfolio(X=X, weights=weights))

        selected = population.sort_measure(measure=self.measure, reverse=self.highest)[
            :k
        ]
        selected_idx = [x.nonzero_assets_index[0] for x in selected]
        self.to_keep_ = np.isin(np.arange(n_assets), selected_idx)
        return self

    def _get_support_mask(self):
        skv.check_is_fitted(self)
        return self.to_keep_


class SelectNonDominated(skf.SelectorMixin, skb.BaseEstimator):
    """Transformer for selecting non dominated assets.

    Pre-selection based on the Assets Preselection Process 2 [1]_.

    Good single asset (for example with high return and low risk) is likely to
    contribute to the final optimized portfolio. Each asset is considered as a portfolio
    and these assets are ranked using the non-domination sorting method. The selection
    is based on the ranks assigned to each asset based on their fitness until the number
    of selected assets reaches the user-defined number.

    Considering only the fitness of individual asset is insufficient because a pair of
    negatively correlated assets has the potential to reduce the risk. Therefore,
    negatively correlated pairs of assets are also considered.

    Parameters
    ----------
    min_n_assets : int, optional
        The minimum number of assets to select. If `min_n_assets` is reached before the
        end of the current non-dominated front, we return the remaining assets of this
        front. This is because all assets in the same front have same rank.
        The default (`None`) is to select the first front.

    threshold : float, default=0.0
        Asset pair with a correlation below this threshold are included in the
        non-domination sorting. The default value is `0.0`.

    fitness_measures : list[Measure], optional
        A list of :ref:`measure <measures_ref>` used to compute the portfolio fitness.
        The fitness is used to compare portfolios in terms of domination, compute the
        pareto fronts and run the portfolio selection using non-denominated sorting.
        The default (`None`) is to use the list [PerfMeasure.MEAN, RiskMeasure.VARIANCE]

    Attributes
    ----------
    to_keep_ : ndarray of shape (n_assets, )
        Boolean array indicating which assets are remaining.

    n_features_in_ : int
        Number of assets seen during `fit`.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of features seen during `fit`. Defined only when `X`
        has feature names that are all strings.

    References
    ----------
    .. [1]  "Large-Scale Portfolio Optimization Using Multi-objective Evolutionary
        Algorithms and Preselection Methods",
        B.Y. Qu and Q.Zhou (2017).
    """

    to_keep_: np.ndarray

    def __init__(
        self,
        min_n_assets: int | None = None,
        threshold: float = -0.5,
        fitness_measures: list[skt.Measure] | None = None,
    ):
        self.min_n_assets = min_n_assets
        self.threshold = threshold
        self.fitness_measures = fitness_measures

    def fit(self, X: npt.ArrayLike, y=None):
        """Run the Non Dominated transformer and get the appropriate assets.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
            Price returns of the assets.

        y : Ignored
            Not used, present for API consistency by convention.

        Returns
        -------
        self : SelectNonDominated
            Fitted estimator.
        """
        X = self._validate_data(X)
        if not -1 <= self.threshold <= 1:
            raise ValueError("`threshold` must be between -1 and 1")
        n_assets = X.shape[1]

        if self.min_n_assets is not None and self.min_n_assets >= n_assets:
            self.to_keep_ = np.full(n_assets, True)
            return self

        # Build a population of portfolio
        population = Population([])
        # Add single assets
        for i in range(n_assets):
            weights = np.zeros(n_assets)
            weights[i] = 1
            population.append(
                Portfolio(X=X, weights=weights, fitness_measures=self.fitness_measures)
            )

        # Add pairs with correlation below threshold with minimum variance
        # ptf_variance = sigma1^2 w1^2 + sigma2^2 w2^2 + 2 sigma12 w1 w2 (1)
        # with w1 + w2 = 1
        # To find the minimum we substitute w2 = 1 - w1 in (1) and differentiate with
        # respect to w1 and set to zero.
        # By solving the obtained equation, we get:
        # w1 = (sigma2^2 - sigma12) / (sigma1^2 + sigma2^2 - 2 sigma12)
        # w2 = 1 - w1

        corr = np.corrcoef(X.T)
        covariance = np.cov(X.T)
        for i, j in zip(*np.triu_indices(n_assets, 1), strict=True):
            if corr[i, j] < self.threshold:
                cov = covariance[i, j]
                var1 = covariance[i, i]
                var2 = covariance[j, j]
                weights = np.zeros(n_assets)
                weights[i] = (var2 - cov) / (var1 + var2 - 2 * cov)
                weights[j] = 1 - weights[i]
                population.append(
                    Portfolio(
                        X=X, weights=weights, fitness_measures=self.fitness_measures
                    )
                )

        fronts = population.non_denominated_sort(
            first_front_only=self.min_n_assets is None
        )
        new_assets_idx = set()
        i = 0
        while i < len(fronts):
            if (
                self.min_n_assets is not None
                and len(new_assets_idx) > self.min_n_assets
            ):
                break
            for idx in fronts[i]:
                new_assets_idx.update(population[idx].nonzero_assets_index)
            i += 1
        self.to_keep_ = np.isin(np.arange(n_assets), list(new_assets_idx))
        return self

    def _get_support_mask(self):
        skv.check_is_fitted(self)
        return self.to_keep_



"""Expected returns estimators."""

# Copyright (c) 2023
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# License: BSD 3 clause
# Implementation derived from:
# Riskfolio-Lib, Copyright (c) 2020-2023, Dany Cajas, Licensed under BSD 3 clause.
# scikit-learn, Copyright (c) 2007-2010 David Cournapeau, Fabian Pedregosa, Olivier
# Grisel Licensed under BSD 3 clause.

from enum import auto

import numpy as np
import numpy.typing as npt
import pandas as pd

from skfolio.moments.covariance import BaseCovariance, EmpiricalCovariance
from skfolio.moments.expected_returns._base import BaseMu
from skfolio.utils.tools import AutoEnum, check_estimator


class EmpiricalMu(BaseMu):
    """Empirical Expected Returns (Mu) estimator.

    Estimates the expected returns with the historical mean.

    Parameters
    ----------
    window_size : int, optional
        Window size. The model is fitted on the last `window_size` observations.
        The default (`None`) is to use all the data.

    Attributes
    ----------
    mu_ : ndarray of shape (n_assets,)
        Estimated expected returns of the assets.

    n_features_in_ : int
        Number of assets seen during `fit`.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of assets seen during `fit`. Defined only when `X`
        has assets names that are all strings.
    """

    def __init__(self, window_size: int | None = None):
        self.window_size = window_size

    def fit(self, X: npt.ArrayLike, y=None) -> "EmpiricalMu":
        """Fit the Mu Empirical estimator model.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
            Price returns of the assets.

        y : Ignored
            Not used, present for API consistency by convention.

        Returns
        -------
        self : EmpiricalMu
            Fitted estimator.
        """
        X = self._validate_data(X)
        if self.window_size is not None:
            X = X[-self.window_size :]
        self.mu_ = np.mean(X, axis=0)
        return self


class EWMu(BaseMu):
    r"""Exponentially Weighted Expected Returns (Mu) estimator.

    Estimates the expected returns with the exponentially weighted mean (EWM).

    Parameters
    ----------
    window_size : int, optional
        Window size. The model is fitted on the last `window_size` observations.
        The default (`None`) is to use all the data.

    alpha : float, default=0.2
        Exponential smoothing factor. The default value is `0.2`.

        :math:`0 < \alpha \leq 1`.

    Attributes
    ----------
    mu_ : ndarray of shape (n_assets,)
       Estimated expected returns of the assets.

    n_features_in_ : int
        Number of assets seen during `fit`.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of assets seen during `fit`. Defined only when `X`
        has assets names that are all strings.
    """

    def __init__(self, window_size: int | None = None, alpha: float = 0.2):
        self.window_size = window_size
        self.alpha = alpha

    def fit(self, X: npt.ArrayLike, y=None) -> "EWMu":
        """Fit the EWMu estimator model.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
           Price returns of the assets.

        y : Ignored
            Not used, present for API consistency by convention.

        Returns
        -------
        self : EWMu
            Fitted estimator.
        """
        X = self._validate_data(X)
        if self.window_size is not None:
            X = X[-self.window_size :]
        self.mu_ = pd.DataFrame(X).ewm(alpha=self.alpha).mean().iloc[-1, :].to_numpy()
        return self


class EquilibriumMu(BaseMu):
    r"""Equilibrium Expected Returns (Mu) estimator.

    The Equilibrium is defined as:

        .. math:: risk\_aversion \times \Sigma \cdot w^T

    For Market Cap Equilibrium, the weights are the assets Market Caps.
    For Equal-weighted Equilibrium, the weights are equal-weighted (1/N).

    Parameters
    ----------
    risk_aversion : float, default=1.0
        Risk aversion factor.
        The default value is `1.0`.

    weights : array-like of shape (n_assets,), optional
        Asset weights used to compute the Expected Return Equilibrium.
        The default is to use the equal-weighted equilibrium (1/N).
        For a Market Cap weighted equilibrium, you must provide the asset Market Caps.

    covariance_estimator : BaseCovariance, optional
        :ref:`Covariance estimator <covariance_estimator>` used to estimate the
        covariance in the equilibrium formula.
        The default (`None`) is to use :class:`~skfolio.moments.EmpiricalCovariance`.

    Attributes
    ----------
    mu_ : ndarray of shape (n_assets,)
          Estimated expected returns of the assets.

    covariance_estimator_ : BaseCovariance
        Fitted `covariance_estimator`.

    n_features_in_ : int
        Number of assets seen during `fit`.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of assets seen during `fit`. Defined only when `X`
        has assets names that are all strings.
    """

    covariance_estimator_: BaseCovariance

    def __init__(
        self,
        risk_aversion: float = 1,
        weights: np.ndarray | None = None,
        covariance_estimator: BaseCovariance | None = None,
    ):
        self.risk_aversion = risk_aversion
        self.weights = weights
        self.covariance_estimator = covariance_estimator

    def fit(self, X: npt.ArrayLike, y=None) -> "EquilibriumMu":
        """Fit the EquilibriumMu estimator model.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
           Price returns of the assets.

        y : Ignored
            Not used, present for API consistency by convention.

        Returns
        -------
        self : EquilibriumMu
            Fitted estimator.
        """
        # fitting estimators
        self.covariance_estimator_ = check_estimator(
            self.covariance_estimator,
            default=EmpiricalCovariance(),
            check_type=BaseCovariance,
        )
        self.covariance_estimator_.fit(X)

        # we validate and convert to numpy after all models have been fitted to keep
        # features names information.
        X = self._validate_data(X)
        n_assets = X.shape[1]
        if self.weights is None:
            weights = np.ones(n_assets) / n_assets
        else:
            weights = np.asarray(self.weights)
        self.mu_ = self.risk_aversion * self.covariance_estimator_.covariance_ @ weights
        return self


class ShrunkMuMethods(AutoEnum):
    """Shrinkage methods for the ShrunkMu estimator

    Parameters
    ----------
    JAMES_STEIN : str
        James-Stein method

    BAYES_STEIN : str
        Bayes-Stein method

    BODNAR_OKHRIN : str
        Bodnar Okhrin Parolya method
    """

    JAMES_STEIN = auto()
    BAYES_STEIN = auto()
    BODNAR_OKHRIN = auto()


class ShrunkMu(BaseMu):
    r"""Shrinkage Expected Returns (Mu) estimator.

    Estimates the expected returns using shrinkage.

    The sample mean estimator is unbiased but has high variance.
    Stein (1955) proved that it's possible to find an estimator with reduced total
    error using shrinkage by trading a small bias against high variance.

    The estimator shrinks the sample mean toward a target vector:

        .. math:: \hat{\mu} = \alpha\bar{\mu}+\beta \mu_{target}

    with :math:`\bar{\mu}` the sample mean, :math:`\mu_{target}` the target vector
    and :math:`\alpha` and :math:`\beta` two constants to determine.

    There are two choices for the target vector :math:`\mu_{target}` :

        * Grand Mean: constant vector of the mean of the sample mean
        * Volatility-Weighted Grand Mean: volatility-weighted sample mean

    And three methods for :math:`\alpha` and :math:`\beta` :

        * James-Stein
        * Bayes-Stein
        * Bodnar Okhrin Parolya

    Parameters
    ----------
    covariance_estimator : BaseCovariance, optional
        :ref:`Covariance estimator <covariance_estimator>` used to estimate the
        covariance in the shrinkage formulae.
        The default (`None`) is to use :class:`~skfolio.moments.EmpiricalCovariance`.

    vol_weighted_target : bool, default=False
        If this is set to True, the target vector :math:`\mu_{target}` is the
        Volatility-Weighted Grand Mean otherwise it is the Grand Mean.
        The default is `False`.

    method : ShrunkMuMethods, default=ShrunkMuMethods.JAMES_STEIN
        Shrinkage method :class:`ShrunkMuMethods`.

        Possible values are:

            * JAMES_STEIN
            * BAYES_STEIN
            * BODNAR_OKHRIN

        The default value is `ShrunkMuMethods.JAMES_STEIN`.

    Attributes
    ----------
    mu_ : ndarray of shape (n_assets,)
        Estimated expected returns of the assets.

    covariance_estimator_ : BaseCovariance
        Fitted `covariance_estimator`.

    mu_target_ : ndarray of shape (n_assets,)
        Target vector :math:`\mu_{target}`.

    alpha_ : float
        Alpha value :math:`\alpha`.

    beta_ : float
        Beta value :math:`\beta`.

    n_features_in_ : int
        Number of assets seen during `fit`.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of assets seen during `fit`. Defined only when `X`
        has assets names that are all strings.

    References
    ----------
    .. [1] "Risk and Asset Allocation",
        Attilio Meucci (2005)

    .. [2] "Bayes-stein estimation for portfolio analysis",
        Philippe Jorion (1986)

    .. [3] "Optimal shrinkage estimator for high-dimensional mean vector"
        Bodnar, Okhrin and Parolya (2019)
    """

    covariance_estimator_: BaseCovariance
    mu_target_: np.ndarray
    alpha_: float
    beta_: float

    def __init__(
        self,
        covariance_estimator: BaseCovariance | None = None,
        vol_weighted_target: bool = False,
        method: ShrunkMuMethods = ShrunkMuMethods.JAMES_STEIN,
    ):
        self.covariance_estimator = covariance_estimator
        self.vol_weighted_target = vol_weighted_target
        self.method = method

    def fit(self, X: npt.ArrayLike, y=None) -> "ShrunkMu":
        """Fit the ShrunkMu estimator model.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
           Price returns of the assets.

        y : Ignored
            Not used, present for API consistency by convention.

        Returns
        -------
        self : ShrunkMu
            Fitted estimator.
        """
        if not isinstance(self.method, ShrunkMuMethods):
            raise ValueError(
                "`method` must be of type ShrunkMuMethods, got"
                f" {type(self.method).__name__}"
            )
        # fitting estimators
        self.covariance_estimator_ = check_estimator(
            self.covariance_estimator,
            default=EmpiricalCovariance(),
            check_type=BaseCovariance,
        )
        self.covariance_estimator_.fit(X)

        # we validate and convert to numpy after all models have been fitted to keep
        # features names information.
        X = self._validate_data(X)
        n_observations, n_assets = X.shape

        covariance = self.covariance_estimator_.covariance_

        sample_mu = np.mean(X, axis=0)
        cov_inv = None

        # Calculate target vector
        if self.vol_weighted_target:
            cov_inv = np.linalg.inv(covariance)
            self.mu_target_ = np.sum(cov_inv, axis=1) @ sample_mu / np.sum(cov_inv)
        else:
            self.mu_target_ = np.mean(sample_mu)
        self.mu_target_ *= np.ones(n_assets)

        # Calculate Estimators
        match self.method:
            case ShrunkMuMethods.JAMES_STEIN:
                eigenvalues = np.linalg.eigvals(covariance)
                self.beta_ = (
                    (np.sum(eigenvalues) - 2 * np.max(eigenvalues))
                    / np.sum((sample_mu - self.mu_target_) ** 2)
                    / n_observations
                )
                self.alpha_ = 1 - self.beta_
            case ShrunkMuMethods.BAYES_STEIN:
                if cov_inv is None:
                    cov_inv = np.linalg.inv(covariance)
                self.beta_ = (n_assets + 2) / (
                    n_observations
                    * (sample_mu - self.mu_target_).T
                    @ cov_inv
                    @ (sample_mu - self.mu_target_)
                    + (n_assets + 2)
                )
                self.alpha_ = 1 - self.beta_
            case ShrunkMuMethods.BODNAR_OKHRIN:
                if cov_inv is None:
                    cov_inv = np.linalg.inv(covariance)
                u = sample_mu.T @ cov_inv @ sample_mu
                v = sample_mu.T @ cov_inv @ self.mu_target_
                w = self.mu_target_.T @ cov_inv @ self.mu_target_
                self.alpha_ = (
                    (u - n_assets / (n_observations - n_assets)) * w - v**2
                ) / (u * w - v**2)
                self.beta_ = (1 - self.alpha_) * v / u
            case _:
                raise ValueError(f"method {self.method} is not valid")

        self.mu_ = self.alpha_ * sample_mu + self.beta_ * self.mu_target_
        return self



"""Hierarchical Clustering estimators."""

# Copyright (c) 2023
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# License: BSD 3 clause

from enum import auto

import numpy as np
import numpy.typing as npt
import plotly.graph_objects as go
import scipy.cluster.hierarchy as sch
import scipy.spatial.distance as scd
import sklearn.base as skb
import sklearn.utils.validation as skv

from skfolio.utils.fixes import create_dendrogram
from skfolio.utils.stats import assert_is_distance, compute_optimal_n_clusters
from skfolio.utils.tools import AutoEnum, default_asset_names


class LinkageMethod(AutoEnum):
    r"""Methods for calculating the distance between clusters in the linkage matrix.
    See the `Linkage Methods` section of `scipy.cluster.hierarchy.linkage`
    for full descriptions.

    Parameters
    ----------
    SINGLE : str
        Assigns

        .. math:: d(u,v) = \min(dist(u[i],v[j]))

        for all points :math:`i` in cluster :math:`u` and
        :math:`j` in cluster :math:`v`. This is also known as the
        Nearest Point Algorithm.

    COMPLETE : str
        Assigns

        .. math:: d(u, v) = \max(dist(u[i],v[j]))

        for all points :math:`i` in cluster u and :math:`j` in
        cluster :math:`v`. This is also known by the Farthest Point
        Algorithm or Voor Hees Algorithm.

    AVERAGE : str
        Assigns

        .. math:: d(u,v) = \sum_{ij} \frac{d(u[i], v[j])}{(|u|*|v|)}

        for all points :math:`i` and :math:`j` where :math:`|u|`
        and :math:`|v|` are the cardinalities of clusters :math:`u`
        and :math:`v`, respectively. This is also called the UPGMA
        algorithm.

    WEIGHTED : str
        Assigns

        .. math:: d(u,v) = (dist(s,v) + dist(t,v))/2

        where cluster u was formed with cluster s and t and v
        is a remaining cluster in the forest (also called WPGMA).

    CENTROID : str
        Assigns

        .. math::
           dist(s,t) = ||c_s-c_t||_2

        where :math:`c_s` and :math:`c_t` are the centroids of
        clusters :math:`s` and :math:`t`, respectively.
        This is also known as the UPGMC
        algorithm.

    MEDIAN : str
    assigns :math:`d(s,t)` like the ``centroid`` method.
    This is also known as the WPGMC algorithm.

    WARD : str
        Uses the Ward variance minimization algorithm.
        The new entry :math:`d(u,v)` is computed as follows,

        .. math::

           d(u,v) = \sqrt{\frac{|v|+|s|}
                               {T}d(v,s)^2
                        + \frac{|v|+|t|}
                               {T}d(v,t)^2
                        - \frac{|v|}
                               {T}d(s,t)^2}

        where :math:`u` is the newly joined cluster consisting of
        clusters :math:`s` and :math:`t`, :math:`v` is an unused
        cluster in the forest, :math:`T=|v|+|s|+|t|`, and
        :math:`|*|` is the cardinality of its argument. This is also
        known as the incremental algorithm.
    """

    SINGLE = auto()
    COMPLETE = auto()
    AVERAGE = auto()
    WEIGHTED = auto()
    CENTROID = auto()
    MEDIAN = auto()
    WARD = auto()


class HierarchicalClustering(skb.ClusterMixin, skb.BaseEstimator):
    r"""Hierarchical Clustering.

    Parameters
    ----------
    max_clusters : int, optional
        For coherent clustering, the algorithm finds a minimum threshold ``r`` so that
        the cophenetic distance between any two original observations in the same flat
        cluster is no more than ``r`` and no more than `max_clusters` flat clusters are
        formed. The default (`None`) is to estimate the maximal number of clusters
        based on the Two-Order Difference to Gap Statistic [1]_.

    linkage_method : LinkageMethod, default=LinkageMethod.WARD
        Methods for calculating the distance between clusters in the linkage matrix.
        See the `Linkage Methods` section of `scipy.cluster.hierarchy.linkage` for
        the full descriptions.
        The default is the Ward variance minimization algorithm `LinkageMethod.WARD`.

    Attributes
    ----------
    n_clusters_ : int
        Number of formed clusters.

    labels_ : ndarray of shape (n_assets,)
        Labels of each asset.

    linkage_matrix_ : ndarray of shape (n_assets - 1, 4)
        Linkage matrix computed from the distance matrix of the `distance_estimator`.

    condensed_distance_ : ndarray of shape (\\binom{n_assets}{2}, )
        The 1-D condensed distance matrix.

    n_features_in_ : int
        Number of assets seen during `fit`.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of assets seen during `fit`. Defined only when `X`
        has assets names that are all strings.

    References
    ----------
    .. [1] "Application of two-order difference to gap statistic".
        Yue, Wang & Wei (2009)
    """

    n_clusters_: int
    labels_: np.ndarray
    linkage_matrix_: np.ndarray
    condensed_distance_: np.ndarray

    def __init__(
        self,
        max_clusters: int | None = None,
        linkage_method: LinkageMethod = LinkageMethod.WARD,
    ):
        self.max_clusters = max_clusters
        self.linkage_method = linkage_method

    def fit(self, X: npt.ArrayLike, y: None = None) -> "HierarchicalClustering":
        """Fit the Hierarchical Equal Risk Contribution estimator.

        Parameters
        ----------
        X : array-like of shape (n_assets, n_assets)
            Distance matrix of the assets.

        y : Ignored
            Not used, present for API consistency by convention.

        Returns
        -------
        self : HierarchicalClustering
            Fitted estimator.
        """
        X = self._validate_data(X)
        assert_is_distance(X)
        self.condensed_distance_ = scd.squareform(X, checks=False)
        self.linkage_matrix_ = sch.linkage(
            self.condensed_distance_,
            method=str(self.linkage_method.value),
            # Not needed for clustering, only for
            # visualization and can be slow. So we perform the optimal ordering only
            # in `plot_dendrogram`.
            optimal_ordering=False,
        )
        max_clusters = self.max_clusters
        if max_clusters is None:
            max_clusters = compute_optimal_n_clusters(
                distance=X,
                linkage_matrix=self.linkage_matrix_,
            )
        # Get the clusters from the linkage matrix
        labels_ = sch.fcluster(
            self.linkage_matrix_, t=max_clusters, criterion="maxclust"
        )
        labels_ -= 1  # Start at 0
        self.n_clusters_ = len(set(labels_))
        assert self.n_clusters_ == max(labels_) + 1 <= max_clusters
        self.labels_ = labels_
        return self

    def plot_dendrogram(self, heatmap: bool = True) -> go.Figure:
        """Plot the dendrogram.

        The blue lines represent distinct clusters composed of a single asset.
        The remaining colors represent clusters of more than one asset.

        When `heatmap` is set to True, the heatmap of the reordered distance matrix is
        displayed below the dendrogram and clusters are outlined with yellow squares.

        The number of clusters used in the plot is the same as the `n_clusters_`
        attribute if it exists, otherwise a default number is used corresponding to the
        number of cluster with a distance above 70% of the maximum cluster distance.

        Parameters
        ----------
        heatmap : bool, default=True
            If this is set to True, the distance heatmap is returned with the clustered
            outlined in yellow.

        Returns
        -------
        fig : Figure
            The dendrogram figure.
        """
        skv.check_is_fitted(self, "linkage_matrix_")
        linkage_matrix = sch.optimal_leaf_ordering(
            self.linkage_matrix_, self.condensed_distance_
        )

        n_assets = linkage_matrix.shape[0] + 1
        cophenetic_distance_threshold = linkage_matrix[-(self.n_clusters_ - 1), 2]

        if hasattr(self, "feature_names_in_"):
            asset_names = self.feature_names_in_
        else:
            asset_names = default_asset_names(n_assets=n_assets)

        if not heatmap:
            fig = create_dendrogram(
                np.ones(1),
                distfun=lambda x: None,
                linkagefun=lambda x: linkage_matrix,
                color_threshold=cophenetic_distance_threshold,
                labels=asset_names,
            )
            fig.update_layout(
                title="Dendrogram",
                width=800,
                height=400,
                showlegend=False,
                hovermode="closest",
                xaxis={"title": "Assets"},
                yaxis={"title": "Distance"},
            )
            return fig

        # Initialize figure by creating upper dendrogram
        fig = create_dendrogram(
            np.ones(1),
            orientation="bottom",
            distfun=lambda x: None,
            linkagefun=lambda x: linkage_matrix,
            color_threshold=cophenetic_distance_threshold,
            labels=asset_names,
        )

        for i in range(len(fig["data"])):
            fig["data"][i]["yaxis"] = "y2"

        # Create Side Dendrogram
        side_dendrogram = create_dendrogram(
            np.ones(1),
            orientation="right",
            distfun=lambda x: None,
            linkagefun=lambda x: linkage_matrix,
            color_threshold=cophenetic_distance_threshold,
            labels=asset_names,
        )
        for i in range(len(side_dendrogram["data"])):
            side_dendrogram["data"][i]["xaxis"] = "x2"

        # Add Side Dendrogram Data to Figure
        for data in side_dendrogram["data"]:
            fig.add_trace(data)

        # Create Heatmap
        ordered_asset_names = side_dendrogram["layout"]["yaxis"]["ticktext"]
        ordered_asset_names_idx = np.array(
            [np.argwhere(x == asset_names)[0][0] for x in ordered_asset_names]
        )
        assert np.array_equal(asset_names[ordered_asset_names_idx], ordered_asset_names)

        distance = scd.squareform(self.condensed_distance_, checks=False)
        heat_data = distance[ordered_asset_names_idx, :][:, ordered_asset_names_idx]

        heatmap = [
            go.Heatmap(
                x=ordered_asset_names,
                y=ordered_asset_names,
                z=heat_data,
                colorscale="Blues",
                name="",
            )
        ]

        heatmap[0]["x"] = fig["layout"]["xaxis"]["tickvals"]
        heatmap[0]["y"] = side_dendrogram["layout"]["yaxis"]["tickvals"]

        # Add Heatmap Data to Figure
        for data in heatmap:
            fig.add_trace(data)

        # Outline clusters
        delta = heatmap[0]["x"][1] - heatmap[0]["x"][0]

        clusters_ids = self.labels_[ordered_asset_names_idx]

        for i in range(max(clusters_ids) + 1):
            c_ids = np.argwhere(clusters_ids == i).ravel()
            a = c_ids[0] * delta
            b = (c_ids[-1] + 1) * delta
            fig.add_shape(
                type="rect",
                x0=a,
                y0=a,
                x1=b,
                y1=b,
                line=dict(
                    color="gold",
                    width=2,
                ),
            )
        fig.update_layout(
            title="Dendrogram",
            width=800,
            height=800,
            showlegend=False,
            hovermode="closest",
            xaxis={
                "title": "Assets",
                "domain": [0.15, 1],
                "mirror": False,
                "showgrid": False,
                "showline": False,
                "zeroline": False,
                "ticks": "",
            },
            xaxis2={
                "domain": [0, 0.15],
                "mirror": False,
                "showgrid": False,
                "showline": False,
                "zeroline": False,
                "showticklabels": False,
                "ticks": "",
            },
            yaxis={
                "title": "Assets",
                "domain": [0, 0.85],
                "mirror": False,
                "showgrid": False,
                "showline": False,
                "zeroline": False,
                "showticklabels": False,
                "ticks": "",
                "tickvals": fig["layout"]["xaxis"]["tickvals"],
                "ticktext": fig["layout"]["xaxis"]["ticktext"],
            },
            yaxis2={
                "domain": [0.825, 0.975],
                "mirror": False,
                "showgrid": False,
                "showline": False,
                "zeroline": False,
                "showticklabels": False,
                "ticks": "",
            },
        )

        return fig



from skfolio.optimization.cluster.hierarchical._base import (
    BaseHierarchicalOptimization,
)
from skfolio.optimization.cluster.hierarchical._herc import (
    HierarchicalEqualRiskContribution,
)
from skfolio.optimization.cluster.hierarchical._hrp import HierarchicalRiskParity

__all__ = [
    "BaseHierarchicalOptimization",
    "HierarchicalRiskParity",
    "HierarchicalEqualRiskContribution",
]



"""Walk Forward cross-validator"""

# Copyright (c) 2023
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# License: BSD 3 clause
# Implementation derived from:
# scikit-portfolio, Copyright (c) 2022, Carlo Nicolini, Licensed under MIT Licence.
# scikit-learn, Copyright (c) 2007-2010 David Cournapeau, Fabian Pedregosa, Olivier
# Grisel Licensed under BSD 3 clause.

from collections.abc import Iterator

import numpy as np
import numpy.typing as npt
import sklearn.model_selection as skm
import sklearn.utils as sku


class WalkForward(skm.BaseCrossValidator):
    """Walk Forward cross-validator.

    Provides train/test indices to split time series data samples in a walk forward
    logic.

    In each split, test indices must be higher than before, and thus shuffling
    in cross validator is inappropriate.

    Compared to `sklearn.model_selection.TimeSeriesSplit`, you control the train/test
    folds by providing a number of training and test samples instead of a number of
    split making it more suitable for portfolio cross-validation.

    Parameters
    ----------
    test_size : int
        Number of observations in each test set.

    train_size : int
        Number of observations in each training set.

    expend_train : bool, default=False
        If this is set to True, each subsequent training set after the first one will
        use all past observations.
        The default is `False`

    reduce_test : bool, default=False
        If this is set to True, the last train/test split will be returned even if the
        test set is partial (if it contains less observations than `test_size`),
        otherwise it will be ignored.
        The default is `False`

    purged_size : int, default=0
        Number of observations to exclude from the end of each train set before the
        test set.
        The default value is `0`

    Examples
    --------
    >>> import numpy as np
    >>> from skfolio.model_selection import WalkForward
    >>> X = np.random.randn(6, 2)
    >>> cv = WalkForward(test_size=1, train_size=2)
    >>> for i, (train_index, test_index) in enumerate(cv.split(X)):
    ...     print(f"Fold {i}:")
    ...     print(f"  Train: index={train_index}")
    ...     print(f"  Test:  index={test_index}")
    Fold 0:
      Train: index=[0 1]
      Test:  index=[2]
    Fold 1:
      Train: index=[1 2]
      Test:  index=[3]
    Fold 2:
      Train: index=[2 3]
      Test:  index=[4]
    Fold 3:
      Train: index=[3 4]
      Test:  index=[5]
    >>> cv = WalkForward(test_size=1, train_size=2, purged_size=1)
    >>> for i, (train_index, test_index) in enumerate(cv.split(X)):
    ...     print(f"Fold {i}:")
    ...     print(f"  Train: index={train_index}")
    ...     print(f"  Test:  index={test_index}")
    Fold 0:
      Train: index=[0 1]
      Test:  index=[3]
    Fold 1:
      Train: index=[1 2]
      Test:  index=[4]
    Fold 2:
      Train: index=[2 3]
      Test:  index=[5]
    >>> cv = WalkForward(test_size=2, train_size=3)
    >>> for i, (train_index, test_index) in enumerate(cv.split(X)):
    ...     print(f"Fold {i}:")
    ...     print(f"  Train: index={train_index}")
    ...     print(f"  Test:  index={test_index}")
    Fold 0:
      Train: index=[0 1 2]
      Test:  index=[3 4]
    >>> cv = WalkForward(test_size=2, train_size=3, reduce_test=True)
    >>> for i, (train_index, test_index) in enumerate(cv.split(X)):
    ...     print(f"Fold {i}:")
    ...     print(f"  Train: index={train_index}")
    ...     print(f"  Test:  index={test_index}")
    Fold 0:
      Train: index=[0 1 2]
      Test:  index=[3 4]
    Fold 1:
      Train: index=[2 3 4]
      Test:  index=[5]
    >>> cv = WalkForward(test_size=2, train_size=3, expend_train=True, reduce_test=True)
    >>> for i, (train_index, test_index) in enumerate(cv.split(X)):
    ...     print(f"Fold {i}:")
    ...     print(f"  Train: index={train_index}")
    ...     print(f"  Test:  index={test_index}")
    Fold 0:
      Train: index=[0 1 2]
      Test:  index=[3 4]
    Fold 1:
      Train: index=[0 1 2 3 4]
      Test:  index=[5]
    """

    def __init__(
        self,
        test_size: int,
        train_size: int,
        expend_train: bool = False,
        reduce_test: bool = False,
        purged_size: int = 0,
    ):
        self.test_size = test_size
        self.train_size = train_size
        self.expend_train = expend_train
        self.reduce_test = reduce_test
        self.purged_size = purged_size

    def split(
        self, X: npt.ArrayLike, y=None, groups=None
    ) -> Iterator[np.ndarray, np.ndarray]:
        """Generate indices to split data into training and test set.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
            Price returns of the assets.

        y : array-like of shape (n_observations, n_targets)
            Always ignored, exists for compatibility.

        groups : array-like of shape (n_observations,)
            Always ignored, exists for compatibility.

        Yields
        ------
        train : ndarray
            The training set indices for that split.

        test : ndarray
            The testing set indices for that split.
        """
        X, y = sku.indexable(X, y)
        n_samples = X.shape[0]
        # Make sure we have enough samples for the given split parameters
        if self.train_size + self.purged_size >= n_samples:
            raise ValueError(
                "The sum of `train_size` with `purged_size` "
                f"({self.train_size + self.purged_size}) cannot be greater than the"
                f" number of samples ({n_samples})."
            )

        indices = np.arange(n_samples)

        test_start = self.train_size + self.purged_size
        while True:
            if test_start >= n_samples:
                return
            test_end = test_start + self.test_size
            train_end = test_start - self.purged_size
            if self.expend_train:
                train_start = 0
            else:
                train_start = train_end - self.train_size

            if test_end > n_samples:
                if not self.reduce_test:
                    return
                yield (
                    indices[train_start:train_end],
                    indices[test_start:],
                )
            else:
                yield (
                    indices[train_start:train_end],
                    indices[test_start:test_end],
                )
            test_start = test_end

    def get_n_splits(self, X: npt.ArrayLike, y=None, groups=None) -> int:
        """Returns the number of splitting iterations in the cross-validator

        Parameters
        ----------
         X : array-like of shape (n_observations, n_assets)
            Price returns of the assets.

        y : array-like of shape (n_observations, n_targets)
            Always ignored, exists for compatibility.

        groups : array-like of shape (n_observations,)
            Always ignored, exists for compatibility.

        Returns
        -------
        n_folds : int
            Returns the number of splitting iterations in the cross-validator.
        """
        if X is None:
            raise ValueError("The 'X' parameter should not be None.")
        X, y = sku.indexable(X, y)
        n_samples = X.shape[0]
        n = n_samples - self.train_size - self.purged_size

        if self.reduce_test and n % self.test_size != 0:
            return n // self.test_size + 1
        return n // self.test_size



"""Population module.
A population is a collection of portfolios.
"""

# Copyright (c) 2023
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# License: BSD 3 clause

import inspect
from typing import Any

import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
import scipy.interpolate as sci

import skfolio.typing as skt
from skfolio.portfolio import BasePortfolio, MultiPeriodPortfolio
from skfolio.utils.sorting import non_denominated_sort
from skfolio.utils.tools import deduplicate_names

pd.options.plotting.backend = "plotly"


class Population(list):
    """Population Class.

    A `Population` is a list of :class:`~skfolio.portfolio.Portfolio` or
    :class:`~skfolio.portfolio.MultiPeriodPortfolio` or both.

    Parameters
    ----------
    iterable : list[BasePortfolio]
        The list of portfolios. Each item can be of type
        :class:`~skfolio.portfolio.Portfolio` and/or
        :class:`~skfolio.portfolio.MultiPeriodPortfolio`.
        Empty list are accepted.
    """

    def __init__(self, iterable: list[BasePortfolio]) -> None:
        super().__init__(self._validate_item(item) for item in iterable)

    def __repr__(self) -> str:
        return "<Population(" + super().__repr__() + ")>"

    def __getitem__(
        self, indices: int | list[int] | slice
    ) -> "BasePortfolio | Population":
        item = super().__getitem__(indices)
        if isinstance(item, list):
            return self.__class__(item)
        return item

    def __setitem__(self, index: int, item: BasePortfolio) -> None:
        super().__setitem__(index, self._validate_item(item))

    def __add__(self, other: BasePortfolio) -> "Population":
        if not isinstance(other, Population):
            raise TypeError(
                f"Cannot add a Population with an object of type {type(other)}"
            )
        return self.__class__(super().__add__(other))

    def insert(self, index, item: BasePortfolio) -> None:
        """Insert portfolio before index."""
        super().insert(index, self._validate_item(item))

    def append(self, item: BasePortfolio) -> None:
        """Append portfolio to the end of the population list."""
        super().append(self._validate_item(item))

    def extend(self, other: BasePortfolio) -> None:
        """Extend population list by appending elements from the iterable."""
        if isinstance(other, type(self)):
            super().extend(other)
        else:
            super().extend(self._validate_item(item) for item in other)

    def set_portfolio_params(self, **params: Any) -> "Population":
        """Set the parameters of all the portfolios.

        Parameters
        ----------
        **params : Any
            Portfolio parameters.

        Returns
        -------
        self : Population
            The Population instance.
        """
        if not params:
            return self
        init_signature = inspect.signature(BasePortfolio.__init__)
        # Consider the constructor parameters excluding 'self'
        valid_params = [
            p.name
            for p in init_signature.parameters.values()
            if p.name != "self" and p.kind != p.VAR_KEYWORD
        ]
        for key in params:
            if key not in valid_params:
                raise ValueError(
                    f"Invalid parameter {key!r} . "
                    f"Valid parameters are: {valid_params!r}."
                )

        for portfolio in self:
            for key, value in params.items():
                setattr(portfolio, key, value)

    @staticmethod
    def _validate_item(
        item: BasePortfolio,
    ) -> BasePortfolio:
        """Validate that items are of type Portfolio or MultiPeriodPortfolio."""
        if isinstance(item, BasePortfolio):
            return item
        raise TypeError(
            "Population only accept items that inherit from BasePortfolio such as "
            "Portfolio or MultiPeriodPortfolio"
            f", got {type(item).__name__}"
        )

    def non_denominated_sort(self, first_front_only: bool = False) -> list[list[int]]:
        """Fast non-dominated sorting.
        Sort the portfolios into different non-domination levels.
        Complexity O(MN^2) where M is the number of objectives and N the number of
        portfolios.

        Parameters
        ----------
        first_front_only : bool, default=False
            If this is set to True, only the first front is sorted and returned.
            The default is `False`.

        Returns
        -------
        fronts : list[list[int]]
            A list of Pareto fronts (lists), the first list includes
            non-dominated portfolios.
        """
        n = len(self)
        if n > 0 and np.any(
            [
                portfolio.fitness_measures != self[0].fitness_measures
                for portfolio in self
            ]
        ):
            raise ValueError(
                "Cannot compute non denominated sorting with Portfolios "
                "containing mixed `fitness_measures`"
            )
        fitnesses = np.array([portfolio.fitness for portfolio in self])
        fronts = non_denominated_sort(
            fitnesses=fitnesses, first_front_only=first_front_only
        )
        return fronts

    def filter(
        self, names: skt.Names | None = None, tags: skt.Tags | None = None
    ) -> "Population":
        """Filter the Population of portfolios by names and tags.
        If both names and tags are provided, the intersection is returned.

        Parameters
        ----------
        names : str | list[str], optional
            If provided, the population is filtered by portfolio names.

        tags :  str | list[str], optional
            If provided, the population is filtered by portfolio tags.

        Returns
        -------
        population : Population
            A new population of portfolios filtered by names and tags.
        """
        if tags is None and names is None:
            return self
        if isinstance(names, str):
            names = [names]
        if isinstance(tags, str):
            tags = [tags]

        if tags is None:
            return self.__class__(
                [portfolio for portfolio in self if portfolio.name in names]
            )
        if names is None:
            return self.__class__(
                [portfolio for portfolio in self if portfolio.tag in tags]
            )
        return self.__class__(
            [
                portfolio
                for portfolio in self
                if portfolio.name in names and portfolio.tag in tags
            ]
        )

    def measures(
        self,
        measure: skt.Measure,
        names: skt.Names | None = None,
        tags: skt.Tags | None = None,
    ) -> np.ndarray:
        """Vector of portfolios measures for each portfolio from the
        population filtered by names and tags.

        Parameters
        ----------
        measure : Measure
            The portfolio measure.

        names :  str | list[str], optional
            If provided, the population is filtered by portfolio names.

        tags : str | list[str], optional
            If provided, the population is filtered by portfolio tags.

        Returns
        -------
        values : ndarray
            The vector of portfolios measures.
        """
        population = self.filter(names=names, tags=tags)
        return np.array([ptf.__getattribute__(measure.value) for ptf in population])

    def measures_mean(
        self,
        measure: skt.Measure,
        names: skt.Names | None = None,
        tags: skt.Tags | None = None,
    ) -> float:
        """Mean of portfolios measures for each portfolio from the
        population filtered by names and tags.

        Parameters
        ----------
        measure : Measure
            The portfolio measure.

        names :  str | list[str], optional
           If provided, the population is filtered by portfolio names.

        tags :  str | list[str], optional
           If provided, the population is filtered by portfolio tags.

        Returns
        -------
        value : float
            The mean of portfolios measures.
        """
        return self.measures(measure=measure, names=names, tags=tags).mean()

    def measures_std(
        self,
        measure: skt.Measure,
        names: skt.Names | None = None,
        tags: skt.Tags | None = None,
    ) -> float:
        """Standard-deviation of portfolios measures for each portfolio from the
        population filtered by names and tags.

        Parameters
        ----------
        measure : Measure
            The portfolio measure.

        names : str | list[str], optional
            If provided, the population is filtered by portfolio names.

        tags : str | list[str], optional
            If provided, the population is filtered by portfolio tags.

        Returns
        -------
        value : float
            The standard-deviation of portfolios measures.
        """
        return self.measures(measure=measure, names=names, tags=tags).std()

    def sort_measure(
        self,
        measure: skt.Measure,
        reverse: bool = False,
        names: skt.Names | None = None,
        tags: skt.Tags | None = None,
    ) -> "Population":
        """Sort the population by a given portfolio measure and filter the portfolios
        by names and tags.

        Parameters
        ----------
        measure : Measure
            The portfolio measure.

        reverse : bool, default=False
            If this is set to True, the order is reversed.

        names : str | list[str], optional
            If provided, the population is filtered by portfolio names.

        tags : str | list[str], optional
            If provided, the population is filtered by portfolio tags.

        Returns
        -------
        values : Populations
            The sorted population.
        """
        population = self.filter(names=names, tags=tags)
        return self.__class__(
            sorted(
                population,
                key=lambda x: x.__getattribute__(measure.value),
                reverse=reverse,
            )
        )

    def quantile(
        self,
        measure: skt.Measure,
        q: float,
        names: skt.Names | None = None,
        tags: skt.Tags | None = None,
    ) -> BasePortfolio:
        """Returns the portfolio corresponding to the `q` quantile for a given portfolio
        measure.

        Parameters
        ----------
        measure : Measure
            The portfolio measure.

        q : float
            The quantile value.

        names : str | list[str], optional
            If provided, the population is filtered by portfolio names.

        tags : str | list[str], optional
            If provided, the population is filtered by portfolio tags.

        Returns
        -------
        values : BasePortfolio
           Portfolio corresponding to the `q` quantile for the measure.
        """
        if not 0 <= q <= 1:
            raise ValueError("The quantile`q` must be between 0 and 1")
        sorted_portfolios = self.sort_measure(
            measure=measure, reverse=False, names=names, tags=tags
        )
        k = max(0, int(np.round(len(sorted_portfolios) * q)) - 1)
        return sorted_portfolios[k]

    def min_measure(
        self,
        measure: skt.Measure,
        names: skt.Names | None = None,
        tags: skt.Tags | None = None,
    ) -> BasePortfolio:
        """Returns the portfolio with the minimum measure.

        Parameters
        ----------
        measure : Measure
            The portfolio measure.

        names : str | list[str], optional
            If provided, the population is filtered by portfolio names.

        tags : str | list[str], optional
            If provided, the population is filtered by portfolio tags.

        Returns
        -------
        values : BasePortfolio
            The portfolio with minimum measure.
        """
        return self.quantile(measure=measure, q=0, names=names, tags=tags)

    def max_measure(
        self,
        measure: skt.Measure,
        names: skt.Names | None = None,
        tags: skt.Tags | None = None,
    ) -> BasePortfolio:
        """Returns the portfolio with the maximum measure.

        Parameters
        ----------
        measure: Measure
            The portfolio measure.

        names : str | list[str], optional
            If provided, the population is filtered by portfolio names.

        tags : str | list[str], optional
            If provided, the population is filtered by portfolio tags.

        Returns
        -------
        values : BasePortfolio
            The portfolio with maximum measure.
        """
        return self.quantile(measure=measure, q=1, names=names, tags=tags)

    def summary(
        self,
        formatted: bool = True,
        names: skt.Names | None = None,
        tags: skt.Tags | None = None,
    ) -> pd.DataFrame:
        """Summary of the portfolios in the population

        Parameters
        ----------
        formatted : bool, default=True
            If this is set to True, the measures are formatted into rounded string with
            units.
            The default is `True`.

        names : str | list[str], optional
            If provided, the population is filtered by portfolio names.

        tags : str | list[str], optional
            If provided, the population is filtered by portfolio tags.

        Returns
        -------
        summary : pandas DataFrame
            The population's portfolios summary
        """

        portfolios = self.filter(names=names, tags=tags)
        df = pd.concat(
            [p.summary(formatted=formatted) for p in portfolios],
            keys=[p.name for p in portfolios],
            axis=1,
        )
        return df

    def composition(
        self,
        names: skt.Names | None = None,
        tags: skt.Tags | None = None,
        display_sub_ptf_name: bool = True,
    ) -> pd.DataFrame:
        """Composition of the portfolios in the population.

        Parameters
        ----------
        names : str | list[str], optional
            If provided, the population is filtered by portfolio names.

        tags : str | list[str], optional
            If provided, the population is filtered by portfolio tags.

        display_sub_ptf_name : bool, default=True
            If this is set to True, each sub-portfolio name composing a multi-period
            portfolio is displayed.

        Returns
        -------
        summary : DataFrame
            Composition of the portfolios in the population.
        """
        portfolios = self.filter(names=names, tags=tags)
        comp_list = []
        for p in portfolios:
            comp = p.composition
            if display_sub_ptf_name:
                if isinstance(p, MultiPeriodPortfolio):
                    comp.rename(
                        columns={c: f"{p.name}_{c}" for c in comp.columns}, inplace=True
                    )
            else:
                comp.rename(columns={c: p.name for c in comp.columns}, inplace=True)
            comp_list.append(comp)

        df = pd.concat(comp_list, axis=1)
        df.columns = deduplicate_names(list(df.columns))
        df.fillna(0, inplace=True)
        return df

    def plot_distribution(
        self,
        measure_list: list[skt.Measure],
        tag_list: list[str] | None = None,
        n_bins: int | None = None,
        **kwargs,
    ) -> go.Figure:
        """Plot the population's distribution for each measure provided in the
        measure list.

        Parameters
        ----------
        measure_list : list[Measure]
            The list of portfolio measures. A different distribution is plotted per
            measure.

        tag_list : list[str], optional
            If this is provided, an additional distribution is plotted per measure
            for each tag provided.

        n_bins : int, optional
            Sets the number of bins.

        Returns
        -------
        plot : Figure
            Returns the plotly Figure object.
        """
        values = []
        labels = []
        for measure in measure_list:
            if tag_list is not None:
                for tag in tag_list:
                    values.append(self.measures(measure=measure, tags=tag))
                    labels.append(f"{measure} - {tag}")
            else:
                values.append(self.measures(measure=measure))
                labels.append(measure.value)

        df = pd.DataFrame(np.array(values).T, columns=labels).melt(
            var_name="Population"
        )
        fig = px.histogram(
            df,
            color="Population",
            barmode="overlay",
            marginal="box",
            nbins=n_bins,
            **kwargs,
        )
        fig.update_layout(title_text="Measures Distribution", xaxis_title="measures")
        return fig

    def plot_cumulative_returns(
        self,
        log_scale: bool = False,
        idx: slice | np.ndarray | None = None,
        names: skt.Names | None = None,
        tags: skt.Tags | None = None,
    ) -> go.Figure:
        """Plot the population's portfolios cumulative returns.
        Non-compounded cumulative returns start at 0.
        Compounded cumulative returns are rescaled to start at 1000.

        Parameters
        ----------
        log_scale : bool, default=False
            If this is set to True, the cumulative returns are displayed with a
            logarithm scale on the y-axis and rebased at 1000. The cumulative returns
            must be compounded otherwise an exception is raise.

        idx : slice | array, optional
            Indexes or slice of the observations to plot.
            The default (`None`) is to take all observations.

        names : str | list[str], optional
            If provided, the population is filtered by portfolio names.

        tags : str | list[str], optional
            If provided, the population is filtered by portfolio tags.

        Returns
        -------
        plot : Figure
            Returns the plot Figure object.
        """
        if idx is None:
            idx = slice(None)
        portfolios = self.filter(names=names, tags=tags)
        if not portfolios:
            raise ValueError("No portfolio found")

        cumulative_returns = []
        names = []
        compounded = []
        for ptf in portfolios:
            cumulative_returns.append(ptf.cumulative_returns_df)
            names.append(f"{ptf.name}_{ptf.tag}" if ptf.tag is not None else ptf.name)
            compounded.append(ptf.compounded)
        compounded = set(compounded)

        if len(compounded) == 2:
            raise ValueError(
                "Some portfolios cumulative returns are compounded while some "
                "are non-compounded. You can change the compounded with"
                "`population.set_portfolio_params(compounded=False)`",
            )
        title = "Cumulative Returns"
        compounded = compounded.pop()
        if compounded:
            yaxis_title = f"{title} (rebased at 1000)"
            if log_scale:
                title = f"{title} (compounded & log scaled)"
            else:
                title = f"{title} (compounded)"
        else:
            if log_scale:
                raise ValueError(
                    "Plotting with logarithm scaling must be done on cumulative "
                    "returns that are compounded as opposed to non-compounded."
                    "You can change to compounded with "
                    "`set_portfolio_params(compounded=True)`"
                )
            yaxis_title = title
            title = f"{title} (non-compounded)"

        df = pd.concat(cumulative_returns, axis=1).iloc[:, idx]
        df.columns = deduplicate_names(names)

        fig = df.plot()
        fig.update_layout(
            title=title,
            xaxis_title="Observations",
            yaxis_title=yaxis_title,
            legend_title_text="Portfolios",
        )
        if compounded:
            fig.update_yaxes(tickformat=".0f")
        else:
            fig.update_yaxes(tickformat=".2%")
        if log_scale:
            fig.update_yaxes(type="log")
        return fig

    def plot_composition(
        self,
        names: skt.Names | None = None,
        tags: skt.Tags | None = None,
        display_sub_ptf_name: bool = True,
    ) -> go.Figure:
        """Plot the compositions of the portfolios in the population.

        Parameters
        ----------
        names : str | list[str], optional
            If provided, the population is filtered by portfolio names.

        tags : str | list[str], optional
            If provided, the population is filtered by portfolio tags.

        display_sub_ptf_name : bool, default=True
            If this is set to True, each sub-portfolio name composing a multi-period
            portfolio is displayed.

        Returns
        -------
        plot : Figure
            Returns the plotly Figure object.
        """
        df = self.composition(
            names=names, tags=tags, display_sub_ptf_name=display_sub_ptf_name
        ).T
        fig = px.bar(df, x=df.index, y=df.columns)
        fig.update_layout(
            title="Portfolios Composition",
            xaxis={
                "title": "Portfolios",
            },
            yaxis={
                "title": "Weight",
                "tickformat": ",.0%",
            },
            legend=dict(yanchor="top", y=0.99, xanchor="left", x=1.15),
        )
        return fig

    def plot_measures(
        self,
        x: skt.Measure,
        y: skt.Measure,
        z: skt.Measure = None,
        to_surface: bool = False,
        hover_measures: list[skt.Measure] | None = None,
        show_fronts: bool = False,
        color_scale: skt.Measure | str | None = None,
        names: skt.Names | None = None,
        tags: skt.Tags | None = None,
        title="Portfolios",
    ) -> go.Figure:
        """Plot the 2D (or 3D) scatter points (or surface) of a given set of
        measures for each portfolio in the population.

        Parameters
        ----------
        x : Measure
            The x-axis measure.

        y : Measure
            The y-axis measure.

        z : Measure, optional
            The z-axis measure.

        to_surface : bool, default=False
            If this is set to True, a surface is estimated.

        hover_measures : list[Measure], optional
            The list of measure to show on point hover.

        show_fronts : bool, default=False
            If this is set to True, the pareto fronts are highlighted.
            The default is `False`.

        color_scale : Measure | str, optional
            If this is provided, a color scale is displayed.

        title : str, default="Portfolios"
            The graph title. The default value is "Portfolios".

        names : str | list[str], optional
            If provided, the population is filtered by portfolio names.

        tags : str | list[str], optional
            If provided, the population is filtered by portfolio tags.

        Returns
        -------
        plot : Figure
            Returns the plotly Figure object.
        """
        portfolios = self.filter(names=names, tags=tags)
        num_fmt = ":.3f"
        hover_data = {x: num_fmt, y: num_fmt, "tag": True}

        if z is not None:
            hover_data[z] = num_fmt

        if hover_measures is not None:
            for measure in hover_measures:
                hover_data[measure] = num_fmt

        columns = list(hover_data)
        columns.append("name")
        if isinstance(color_scale, skt.Measure):
            hover_data[color_scale] = num_fmt

        if color_scale is not None and color_scale not in columns:
            columns.append(color_scale)

        col_values = [e.value if isinstance(e, skt.Measure) else e for e in columns]
        res = [
            [portfolio.__getattribute__(attr) for attr in col_values]
            for portfolio in portfolios
        ]
        # Improved formatting
        columns = [str(e) for e in columns]
        hover_data = {str(k): v for k, v in hover_data.items()}

        df = pd.DataFrame(res, columns=columns)
        df["tag"] = df["tag"].astype(str).replace("None", "")

        if show_fronts:
            fronts = self.non_denominated_sort(first_front_only=False)
            if tags is not None:
                ValueError("Cannot plot front with tags selected")
            df["front"] = str(-1)
            for i, front in enumerate(fronts):
                for idx in front:
                    df.iloc[idx, -1] = str(i)
            color = df.columns[-1]
        elif color_scale is not None:
            color = str(color_scale)
        else:
            color = "tag"

        if z is not None:
            if to_surface:
                # estimate the surface
                x_arr = np.array(df[str(x)])
                y_arr = np.array(df[str(y)])
                z_arr = np.array(df[str(z)])

                xi = np.linspace(start=min(x_arr), stop=max(x_arr), num=100)
                yi = np.linspace(start=min(y_arr), stop=max(y_arr), num=100)

                X, Y = np.meshgrid(xi, yi)
                Z = sci.griddata(
                    points=(x_arr, y_arr), values=z_arr, xi=(X, Y), method="cubic"
                )
                fig = go.Figure(
                    go.Surface(
                        x=xi,
                        y=yi,
                        z=Z,
                        hovertemplate="<br>".join(
                            [
                                str(e)
                                + ": %{"
                                + v
                                + ":"
                                + (",.3%" if not e.is_ratio else None)
                                + "}"
                                for e, v in [(x, "x"), (y, "y"), (z, "z")]
                            ]
                        )
                        + "<extra></extra>",
                        colorbar=dict(
                            title=str(z),
                            titleside="top",
                            tickformat=",.2%" if not z.is_ratio else None,
                        ),
                    )
                )

                fig.update_layout(
                    title=title,
                    scene=dict(
                        xaxis={
                            "title": str(x),
                            "tickformat": ",.1%" if not x.is_ratio else None,
                        },
                        yaxis={
                            "title": str(y),
                            "tickformat": ",.1%" if not y.is_ratio else None,
                        },
                        zaxis={
                            "title": str(z),
                            "tickformat": ",.1%" if not z.is_ratio else None,
                        },
                    ),
                )
            else:
                # plot the points
                fig = px.scatter_3d(
                    df,
                    x=str(x),
                    y=str(y),
                    z=str(z),
                    hover_name="name",
                    hover_data=hover_data,
                    color=color,
                    symbol="tag",
                )
                fig.update_traces(marker_size=8)
                fig.update_layout(
                    title=title,
                    scene=dict(
                        xaxis={
                            "title": str(x),
                            "tickformat": ",.1%" if not x.is_ratio else None,
                        },
                        yaxis={
                            "title": str(y),
                            "tickformat": ",.1%" if not y.is_ratio else None,
                        },
                        zaxis={
                            "title": str(z),
                            "tickformat": ",.1%" if not z.is_ratio else None,
                        },
                    ),
                    legend=dict(yanchor="top", y=0.99, xanchor="left", x=1.15),
                )

        else:
            fig = px.scatter(
                df,
                x=str(x),
                y=str(y),
                hover_name="name",
                hover_data=hover_data,
                color=color,
                symbol="tag",
            )
            fig.update_traces(marker_size=10)
            fig.update_layout(
                title=title,
                xaxis={
                    "title": str(x),
                    "tickformat": ",.1%" if not x.is_ratio else None,
                },
                yaxis={
                    "title": str(y),
                    "tickformat": ",.1%" if not y.is_ratio else None,
                },
                legend=dict(yanchor="top", y=0.96, xanchor="left", x=1.25),
            )
        return fig
